# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Basic and baseline pipeline components.
"""

from .bias import BiasModel, BiasScorer, Damping
from .candidates import AllTrainingItemsCandidateSelector, UnratedTrainingItemsCandidateSelector
from .composite import FallbackScorer
from .history import UserTrainingHistoryLookup
from .popularity import PopScorer
from .random import RandomSelector, SoftmaxRanker
from .topn import TopNRanker

__all__ = [
    "BiasModel",
    "BiasScorer",
    "Damping",
    "PopScorer",
    "TopNRanker",
    "RandomSelector",
    "SoftmaxRanker",
    "UserTrainingHistoryLookup",
    "UnratedTrainingItemsCandidateSelector",
    "AllTrainingItemsCandidateSelector",
    "FallbackScorer",
]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Per-user rating holdout methods for user-based data splitting.
"""

from typing import Protocol

import numpy as np

from lenskit.data import ItemList
from lenskit.random import RNGInput, random_generator


class HoldoutMethod(Protocol):
    """
    Holdout methods select test rows for a user (or occasionally an item).
    Partition methods are callable; when called with a data frame, they return
    the test entries.

    Stability:
        Caller
    """

    def __call__(self, items: ItemList) -> ItemList:
        """
        Subset an item list (in the uncommon case of item-based holdouts, the
        item list actually holds user IDs).

        Args:
            udf:
                The item list from which holdout items should be selected.

        Returns:
            The list of test items.
        """
        raise NotImplementedError()


class SampleN(HoldoutMethod):
    """
    Randomly select a fixed number of test rows per user/item.

    Stability:
        Caller

    Args:
        n:
            The number of test items to select.
        rng:
            The random number generator or seed (see :ref:`rng`).
    """

    n: int
    rng: np.random.Generator

    def __init__(self, n: int, rng: RNGInput = None):
        self.n = n
        self.rng = random_generator(rng)

    def __call__(self, items: ItemList) -> ItemList:
        if len(items) <= self.n:
            return items

        sel = self.rng.choice(len(items), self.n, replace=False)
        return items[sel]


class SampleFrac(HoldoutMethod):
    """
    Randomly select a fraction of test rows per user/item.

    Stability:
        Caller

    Args:
        frac:
            The fraction items to select for testing.
        rng:
            The random number generator or seed (see :ref:`rng`).
    """

    fraction: float
    rng: np.random.Generator

    def __init__(self, frac: float, rng: RNGInput = None):
        self.fraction = frac
        self.rng = random_generator(rng)

    def __call__(self, items: ItemList) -> ItemList:
        n = round(len(items) * self.fraction)
        sel = self.rng.choice(len(items), n, replace=False)
        return items[sel]


class LastN(HoldoutMethod):
    """
    Select a fixed number of test rows per user/item, based on ordering by a
    field.

    Stability:
        Caller

    Args:
        n: The number of test items to select.
        field: The field to order by.
    """

    n: int
    field: str

    def __init__(self, n: int, field: str = "timestamp"):
        self.n = n
        self.field = field

    def __call__(self, items: ItemList) -> ItemList:
        if len(items) <= self.n:
            return items

        col = items.field(self.field)
        if col is None:
            raise TypeError(f"item list does not have ordering field {self.field}")
        ordered = np.argsort(col)
        return items[ordered[-self.n :]]


class LastFrac(HoldoutMethod):
    """
    Select a fraction of test rows per user/item.

    Stability:
        Caller

    Args:
        frac(double): the fraction of items to select for testing.
    """

    fraction: float
    field: str

    def __init__(self, frac: float, field: str = "timestamp"):
        self.fraction = frac
        self.field = field

    def __call__(self, items: ItemList) -> ItemList:
        n = round(len(items) * self.fraction)

        col = items.field(self.field)
        if col is None:
            raise TypeError(f"item list does not have ordering field {self.field}")
        ordered = np.argsort(col)
        return items[ordered[-n:]]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from importlib.metadata import entry_points

import click

from lenskit import __version__
from lenskit.logging import LoggingConfig, console

__all__ = ["lenskit", "main", "version"]


def main():
    """
    Run the main LensKit CLI.  This just delegates to :fun:`lenskit`, but pretty-prints errors.
    """
    try:
        lenskit()
    except Exception as e:
        console.print(e)


@click.group("lenskit")
@click.option("-v", "--verbose", "verbosity", count=True, help="enable verbose logging output")
def lenskit(verbosity: int):
    """
    Data and pipeline operations with LensKit.
    """
    lc = LoggingConfig()
    if verbosity:
        lc.set_verbose(verbosity)
    lc.apply()


@lenskit.command("version")
def version():
    """
    Print LensKit version info.
    """
    console.print(f"LensKit version [bold cyan]{__version__}[/bold cyan].")


cli_plugins = entry_points(group="lenskit.cli.plugins")
for plugin in cli_plugins:
    cmd = plugin.load()
    lenskit.add_command(cmd)



"""
Item list collections.
"""

from ._base import ItemListCollection, ItemListCollector, MutableItemListCollection
from ._keys import GenericKey, UserIDKey
from ._list import ListILC

__all__ = [
    "GenericKey",
    "UserIDKey",
    "ItemListCollection",
    "ItemListCollector",
    "MutableItemListCollection",
    "ListILC",
]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Data containers, the internal storage of data sets.
"""

from dataclasses import dataclass
from os import PathLike
from pathlib import Path
from shutil import rmtree

import pyarrow as pa
from pyarrow.parquet import read_table, write_table

from lenskit.logging import get_logger

from .schema import DataSchema

_log = get_logger(__name__)


@dataclass(eq=False, order=False)
class DataContainer:
    schema: DataSchema
    tables: dict[str, pa.Table]

    def save(self, path: str | PathLike[str]):
        """
        Save the data to disk.
        """
        from .summary import save_stats

        path = Path(path)
        log = _log.bind(name=self.schema.name, path=str(path))
        log.info("saving dataset")

        if path.exists():
            log.warn("path already exists, removing")
            rmtree(path)

        log.debug("ensuring path exists")
        path.mkdir(exist_ok=True, parents=True)

        log.debug("writing schema")
        with open(path / "schema.json", "wt") as jsf:
            print(self.schema.model_dump_json(), file=jsf)

        for name, table in self.tables.items():
            log.debug("writing table", table=name, rows=table.num_rows)
            write_table(table, path / f"{name}.parquet", compression="zstd")

        log.debug("writing summary file")
        save_stats(self, path / "summary.md")

    @classmethod
    def load(cls, path: str | PathLike[str]):
        """
        Load data from disk.
        """
        path = Path(path)
        log = _log.bind(path=str(path))

        log.info("loading dataset")
        log.debug("reading schema")
        schema_file = path / "schema.json"
        schema = DataSchema.model_validate_json(schema_file.read_text(encoding="utf8"))
        log = log.bind(name=schema.name)

        tables = {}
        for name in schema.entities:
            log.debug("reading entity table", table=name)
            tables[name] = read_table(path / f"{name}.parquet")

        for name in schema.relationships:
            log.debug("reading relationship table", table=name)
            tables[name] = read_table(path / f"{name}.parquet")

        return cls(schema, tables)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Matrix layouts.
"""

# pyright: basic
from __future__ import annotations

from typing import NamedTuple, TypeVar

import numpy as np
import scipy.sparse as sps
import torch

t = torch
M = TypeVar("M", "CSRStructure", sps.csr_array, sps.coo_array, sps.spmatrix, t.Tensor)


class CSRStructure(NamedTuple):
    """
    Representation of the compressed sparse row structure of a sparse matrix,
    without any data values.

    Stability:
        Caller
    """

    rowptrs: np.ndarray
    colinds: np.ndarray
    shape: tuple[int, int]

    @property
    def nrows(self):
        return self.shape[0]

    @property
    def ncols(self):
        return self.shape[1]

    @property
    def nnz(self):
        return self.rowptrs[self.nrows]

    def extent(self, row: int) -> tuple[int, int]:
        return self.rowptrs[row], self.rowptrs[row + 1]

    def row_cs(self, row: int) -> np.ndarray:
        sp, ep = self.extent(row)
        return self.colinds[sp:ep]


class COOStructure(NamedTuple):
    """
    Representation of the coordinate structure of a sparse matrix, without any
    data values.

    Stability:
        Caller
    """

    row_numbers: np.ndarray
    col_numbers: np.ndarray
    shape: tuple[int, int]

    @property
    def nrows(self):
        return self.shape[0]

    @property
    def ncols(self):
        return self.shape[1]

    @property
    def nnz(self):
        return self.row_numbers[self.nrows]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging

from lenskit.data import Dataset
from lenskit.pipeline import Component, RecPipelineBuilder
from lenskit.random import RNGInput

from .bulk import RunAnalysis, RunAnalysisResult
from .predict import MAE, RMSE
from .ranking import NDCG, RBP, Hit, Recall, RecipRank

_log = logging.getLogger(__name__)


def quick_measure_model(
    model: Component,
    data: Dataset,
    *,
    predicts_ratings: bool = False,
    n_jobs: int | None = 1,
    rng: RNGInput = None,
) -> RunAnalysisResult:
    """
    Do a quick-and-dirty model measurement with a default pipeline setup, split,
    and metrics. This is mostly to make tests easy to write, you usually don't
    want to use it for actual recommender evaluation.

    Stability:
        Caller
    """
    from lenskit.basic import BiasScorer
    from lenskit.batch import BatchPipelineRunner
    from lenskit.splitting import SampleFrac, sample_users

    builder = RecPipelineBuilder()
    builder.scorer(model)
    if predicts_ratings:
        builder.predicts_ratings(fallback=BiasScorer())

    pipe = builder.build()

    n_users = data.user_count
    us_size = n_users // 5
    split = sample_users(data, us_size, SampleFrac(0.2, rng=rng), rng=rng)
    _log.info("measuring %s on %d users", model, us_size)

    pipe.train(split.train)
    runner = BatchPipelineRunner(n_jobs=n_jobs)
    runner.recommend(n=20)
    if predicts_ratings:
        runner.predict()

    outs = runner.run(pipe, split.test)

    rra = RunAnalysis()
    rra.add_metric(RecipRank())
    rra.add_metric(RBP())
    rra.add_metric(NDCG())
    rra.add_metric(Hit())
    rra.add_metric(Recall())

    result = rra.measure(outs.output("recommendations"), split.test)

    if predicts_ratings:
        pra = RunAnalysis()
        pra.add_metric(RMSE())
        pra.add_metric(MAE())
        pr = pra.measure(outs.output("predictions"), split.test)
        result.merge_from(pr)

    return result



import os
import platform
import re
import sys
from dataclasses import dataclass
from importlib.metadata import distributions, version
from pathlib import Path

import click
import threadpoolctl
from humanize import naturalsize
from rich.console import Console, ConsoleOptions, group
from rich.padding import Padding
from rich.table import Table

from lenskit import __version__
from lenskit.logging import get_logger, stdout_console
from lenskit.parallel import ensure_parallel_init
from lenskit.parallel.ray import ray_available

_log = get_logger(__name__)
_gh_out: Path | None = None


@click.command("doctor")
@click.option(
    "--github-output",
    "gh_output",
    envvar="GITHUB_OUTPUT",
    type=Path,
    help="Path to GitHub Actions output file.",
)
@click.option("--packages/--no-packages", default=True, help="List installed packages.")
def doctor(gh_output: Path | None, packages: bool):
    """
    Inspect installed LensKit version and environment.
    """
    global _gh_out
    _gh_out = gh_output
    ensure_parallel_init()
    console = stdout_console()
    console.print(inspect_version())
    console.print(inspect_platform())
    console.print(inspect_compute())
    if ray_available():
        console.print(inspect_ray())
    console.print(inspect_env())
    if packages:
        console.print(inspect_packages())


@dataclass
class kvp:
    name: str
    value: str | int | bool | float | None
    level: int = 1
    value_style: str = "cyan"

    def __rich_console__(self, console: Console, options: ConsoleOptions):
        text = ""
        if self.level == 1:
            text += f"[bold]{self.name}[/bold]: "
        else:
            text += "  " * (self.level - 1)
            text += self.name + ": "
        text += f"[{self.value_style}]"
        text += str(self.value)
        text += f"[/{self.value_style}]"
        return [text]


@group()
def inspect_version():
    dist_ver = version("lenskit")
    if _gh_out:
        with _gh_out.open("at") as ghf:
            print(f"lenskit_version={dist_ver}", file=ghf)

    yield f"[bold]LensKit version[/bold] [cyan]{dist_ver}[/cyan]"
    if str(dist_ver) != __version__:
        yield f"   [yellow]Version mismatch, internal package version is {__version__}[/yellow]"


@group()
def inspect_platform():
    yield kvp("Python version", platform.python_version())
    yield kvp("Platform", platform.platform(), level=2)
    yield kvp("Location", sys.executable, level=2)


@group()
def inspect_compute():
    import numpy as np
    import torch

    yield ""
    yield kvp("NumPy version", np.__version__)
    yield kvp("PyTorch version", torch.__version__)
    if _gh_out:
        with _gh_out.open("at") as ghf:
            print(f"numpy_version={np.__version__}", file=ghf)
            print(f"pytorch_version={torch.__version__}", file=ghf)

    yield "[bold]PyTorch backends[/bold]:"
    yield kvp("cpu", torch.backends.cpu.get_cpu_capability(), level=2)
    for mod in [torch.cuda, torch.backends.mkl, torch.backends.mps]:
        if mod.is_available():
            stat = "available"
        elif hasattr(mod, "is_built") and mod.is_built():
            stat = "unavailable"
        else:
            stat = "absent"
        name = mod.__name__.split(".")[-1]
        yield kvp(name, stat, level=2)

    yield ""
    yield "[bold]Threading layers[/bold]:"
    for i, pool in enumerate(threadpoolctl.threadpool_info(), 1):
        yield f"  Backend {i}:"
        for k, v in pool.items():
            yield kvp(k, v, level=3)


@group()
def inspect_ray():
    import ray

    yield ""
    yield "[bold]Ray cluster ([yellow]experimental[/yellow])[/bold]:"

    try:
        ray.init("auto", configure_logging=False)
    except ConnectionError:
        yield "  Installed but inactive"
    else:
        yield "  Resources:"
        for name, val in ray.cluster_resources().items():
            if name.startswith("node:"):
                continue
            if name.endswith("memory"):
                val = naturalsize(val)
            yield kvp(name, val, level=3)


@group()
def inspect_env():
    yield ""
    yield "[bold]Relevant environment variables[/bold]:"
    for k, v in os.environ.items():
        if re.match(r"^(LK_|OMP_|NUMBA_|MKL_|TORCH_|PY)", k):
            yield kvp(k, v, level=2)

    yield ""
    yield "[bold]Python search paths[/bold]:"
    for path in sys.path:
        yield f"- {path}"

    yield ""
    yield "[bold]Executable search paths[/bold]:"
    paths = os.environ["PATH"].split(os.pathsep)
    for path in paths:
        yield f"- {path}"


def inspect_packages():
    dists = sorted(distributions(), key=lambda d: d.name)

    n = len(dists)
    table = Table(title=f"Installed Packages ({n})")
    table.add_column("Package")
    table.add_column("Version", justify="right")

    for dist in dists:
        table.add_row(dist.name, dist.version)

    return Padding(table, (1, 0, 0, 2))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Console and related logging support
"""

import atexit
import sys
from logging import Handler, LogRecord

from rich.ansi import AnsiDecoder
from rich.console import Console
from rich.live import Live

console = Console(stderr=True)
_live: Live | None = None


class ConsoleHandler(Handler):
    """
    Lightweight Rich log handler for routing StructLog-formatted logs.
    """

    _decoder = AnsiDecoder()

    @property
    def supports_color(self) -> bool:
        return (console.is_terminal or console.is_jupyter) and not console.no_color

    def emit(self, record: LogRecord) -> None:
        try:
            fmt = self.format(record)
            console.print(self._decoder.decode_line(fmt))
        except Exception:
            self.handleError(record)


def get_live() -> Live | None:
    return _live


def setup_console():
    global _live
    if _live is not None:
        return
    if not console.is_terminal:
        return

    _live = Live(console=console, transient=True, redirect_stdout=sys.stdout.isatty())
    _live.start()


def stdout_console():
    """
    Get a console attached to ``stdout``.
    """
    if sys.stdout.isatty():
        return console
    else:
        return Console(stderr=False)


@atexit.register
def _stop_console():
    if _live is not None:
        _live.stop()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
FunkSVD (biased MF).
"""

import logging
import time

import numba as n
import numpy as np
from numba.experimental import jitclass
from pydantic import BaseModel
from typing_extensions import override

from lenskit.basic import BiasModel, Damping
from lenskit.data import Dataset, ItemList, QueryInput, RecQuery, Vocabulary
from lenskit.logging import Stopwatch
from lenskit.pipeline import Component
from lenskit.training import Trainable, TrainingOptions

_logger = logging.getLogger(__name__)


class FunkSVDConfig(BaseModel):
    "Configuration for :class:`FunkSVDScorer`."

    features: int = 50
    """
    Number of latent features.
    """
    epochs: int = 100
    """
    Number of training epochs (per feature).
    """
    learning_rate: float = 0.001
    """
    Gradient descent learning rate.
    """
    regularization: float = 0.015
    """
    Parameter regularization.
    """
    damping: Damping = 5.0
    """
    Bias damping term.
    """
    range: tuple[float, float] | None = None
    """
    Min/max range of ratings to clamp output.
    """


@jitclass(
    [
        ("user_features", n.float64[:, :]),
        ("item_features", n.float64[:, :]),
        ("feature_count", n.int32),
        ("user_count", n.int32),
        ("item_count", n.int32),
        ("initial_value", n.float64),
    ]
)  # type: ignore
class Model:
    "Internal model class for training SGD MF."

    def __init__(self, umat, imat):
        self.user_features = umat
        self.item_features = imat
        self.feature_count = umat.shape[1]
        assert imat.shape[1] == self.feature_count
        self.user_count = umat.shape[0]
        self.item_count = imat.shape[0]
        self.initial_value = np.nan


def _fresh_model(nfeatures, nusers, nitems, init=0.1):
    umat = np.full([nusers, nfeatures], init, dtype=np.float64)
    imat = np.full([nitems, nfeatures], init, dtype=np.float64)
    model = Model(umat, imat)
    model.initial_value = init
    assert model.feature_count == nfeatures
    assert model.user_count == nusers
    assert model.item_count == nitems
    return model


@jitclass(
    [
        ("iter_count", n.int32),
        ("lrate", n.float64),
        ("reg_term", n.float64),
        ("rmin", n.float64),
        ("rmax", n.float64),
    ]
)  # type: ignore
class _Params:
    def __init__(self, niters, lrate, reg, rmin, rmax):
        self.iter_count = niters
        self.lrate = lrate
        self.reg_term = reg
        self.rmin = rmin
        self.rmax = rmax


def make_params(niters, lrate, reg, range):
    if range is None:
        rmin = -np.inf
        rmax = np.inf
    else:
        rmin, rmax = range

    return _Params(niters, lrate, reg, rmin, rmax)


@jitclass([("est", n.float64[:]), ("feature", n.int32), ("trail", n.float64)])  # type: ignore
class _FeatContext:
    def __init__(self, est, feature, trail):
        self.est = est
        self.feature = feature
        self.trail = trail


@jitclass(
    [
        ("users", n.int32[:]),
        ("items", n.int32[:]),
        ("ratings", n.float64[:]),
        ("bias", n.float64[:]),
        ("n_samples", n.uint64),
    ]
)  # type: ignore
class Context:
    def __init__(self, users, items, ratings, bias):
        self.users = users
        self.items = items
        self.ratings = ratings
        self.bias = bias
        self.n_samples = users.shape[0]

        assert items.shape[0] == self.n_samples
        assert ratings.shape[0] == self.n_samples
        assert bias.shape[0] == self.n_samples


@n.njit
def _feature_loop(ctx: Context, params: _Params, model: Model, fc: _FeatContext):
    users = ctx.users
    items = ctx.items
    ratings = ctx.ratings
    umat = model.user_features
    imat = model.item_features
    est = fc.est
    f = fc.feature
    trail = fc.trail

    sse = 0.0
    acc_ud = 0.0
    acc_id = 0.0
    for s in range(ctx.n_samples):
        user = users[s]
        item = items[s]
        ufv = umat[user, f]
        ifv = imat[item, f]

        pred = est[s] + ufv * ifv + trail
        if pred < params.rmin:
            pred = params.rmin
        elif pred > params.rmax:
            pred = params.rmax

        error = ratings[s] - pred
        sse += error * error

        # compute deltas
        ufd = error * ifv - params.reg_term * ufv
        ufd = ufd * params.lrate
        acc_ud += ufd * ufd
        ifd = error * ufv - params.reg_term * ifv
        ifd = ifd * params.lrate
        acc_id += ifd * ifd
        umat[user, f] += ufd
        imat[item, f] += ifd

    return np.sqrt(sse / ctx.n_samples)


@n.njit
def _train_feature(ctx, params, model, fc):
    for epoch in range(params.iter_count):
        rmse = _feature_loop(ctx, params, model, fc)

    return rmse


def train(ctx: Context, params: _Params, model: Model, timer):
    est = ctx.bias

    for f in range(model.feature_count):
        start = time.perf_counter()
        trail = model.initial_value * model.initial_value * (model.feature_count - f - 1)
        fc = _FeatContext(est, f, trail)
        rmse = _train_feature(ctx, params, model, fc)
        end = time.perf_counter()
        _logger.info("[%s] finished feature %d (RMSE=%f) in %.2fs", timer, f, rmse, end - start)

        est = est + model.user_features[ctx.users, f] * model.item_features[ctx.items, f]
        est = np.maximum(est, params.rmin)
        est = np.minimum(est, params.rmax)


def _align_add_bias(bias, index, keys, series):
    "Realign a bias series with an index, and add to a series"
    # realign bias to make sure it matches
    bias = bias.reindex(index, fill_value=0)
    assert len(bias) == len(index)
    # look up bias for each key
    ibs = bias.loc[keys]
    # change index
    ibs.index = keys.index
    # and add
    series = series + ibs
    return bias, series


class FunkSVDScorer(Trainable, Component[ItemList]):
    """
    FunkSVD explicit-feedback matrix factoriation.  FunkSVD is a regularized
    biased matrix factorization technique trained with featurewise stochastic
    gradient descent.

    See the base class :class:`.MFPredictor` for documentation on the estimated
    parameters you can extract from a trained model.

    .. deprecated:: LKPY

        This scorer is kept around for historical comparability, but ALS
        :class:`~lenskit.als.BiasedMF` is usually a better option.

    Stability:
        Caller
    """

    config: FunkSVDConfig

    bias_: BiasModel
    users_: Vocabulary
    user_features_: np.ndarray[tuple[int, int], np.dtype[np.float64]]
    items_: Vocabulary
    item_features_: np.ndarray[tuple[int, int], np.dtype[np.float64]]

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        """
        Train a FunkSVD model.

        Args:
            ratings: the ratings data frame.
        """
        if hasattr(self, "item_features_") and not options.retrain:
            return

        timer = Stopwatch()
        rate_df = data.interaction_matrix(format="pandas", layout="coo", field="rating")

        _logger.info("[%s] fitting bias model", timer)
        self.bias_ = BiasModel.learn(data, damping=self.config.damping)

        _logger.info("[%s] preparing rating data for %d samples", timer, len(rate_df))
        _logger.debug("shuffling rating data")
        shuf = np.arange(len(rate_df), dtype=np.int_)
        rng = options.random_generator()
        rng.shuffle(shuf)
        rate_df = rate_df.iloc[shuf, :]

        users = np.array(rate_df["user_num"])
        items = np.array(rate_df["item_num"])
        ratings = np.array(rate_df["rating"], dtype=np.float64)

        _logger.debug("[%s] computing initial estimates", timer)
        initial = np.full(len(users), self.bias_.global_bias, dtype=np.float64)
        if self.bias_.item_biases is not None:
            initial += self.bias_.item_biases[items]
        if self.bias_.user_biases is not None:
            initial += self.bias_.user_biases[users]

        _logger.debug("have %d estimates for %d ratings", len(initial), len(rate_df))
        assert len(initial) == len(rate_df)

        _logger.debug("[%s] initializing data structures", timer)
        context = Context(users, items, ratings, initial)
        params = make_params(
            self.config.epochs,
            self.config.learning_rate,
            self.config.regularization,
            self.config.range,
        )

        model = _fresh_model(self.config.features, data.users.size, data.items.size)

        _logger.info("[%s] training biased MF model with %d features", timer, self.config.features)
        train(context, params, model, timer)
        _logger.info("finished model training in %s", timer)

        self.users_ = data.users
        self.items_ = data.items
        self.user_features_ = model.user_features
        self.item_features_ = model.item_features

    @override
    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        query = RecQuery.create(query)

        user_id = query.user_id
        user_num = None
        if user_id is not None:
            user_num = self.users_.number(user_id, missing=None)
        if user_num is None:
            _logger.debug("unknown user %s", query.user_id)
            return ItemList(items, scores=np.nan)

        u_feat = self.user_features_[user_num, :]

        item_nums = items.numbers(vocabulary=self.items_, missing="negative")
        item_mask = item_nums >= 0
        i_feats = self.item_features_[item_nums[item_mask], :]

        scores = np.full((len(items),), np.nan, dtype=np.float64)
        scores[item_mask] = i_feats @ u_feat
        biases, _ub = self.bias_.compute_for_items(items, user_id, query.user_items)
        scores += biases

        return ItemList(items, scores=scores)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Logging pipeline configuration.
"""

from __future__ import annotations

import logging
import os
import re
import sys
import warnings
from pathlib import Path
from typing import Literal, TypeAlias

import structlog

from ._console import ConsoleHandler, console, setup_console
from .processors import format_timestamp, log_warning, remove_internal
from .progress import set_progress_impl
from .tracing import lenskit_filtering_logger

LVL_TRACE = 5
CORE_PROCESSORS = [
    structlog.processors.add_log_level,
    structlog.stdlib.add_logger_name,
    structlog.stdlib.PositionalArgumentsFormatter(),
    structlog.processors.MaybeTimeStamper(),
]
LogFormat: TypeAlias = Literal["json", "logfmt", "text"]

_active_config: LoggingConfig | None = None


def active_logging_config() -> LoggingConfig | None:
    """
    Get the currently-active logging configuration.

    Stability:
        Internal
    """
    return _active_config


def basic_logging(level: int = logging.INFO):
    """
    Simple one-function logging configuration for simple command lines.

    Stability:
        Caller
    """
    cfg = LoggingConfig()
    cfg.level = level
    cfg.apply()


def notebook_logging(level: int = logging.INFO):
    """
    Simple one-function logging configuration for notebooks and similar.

    Stability:
        Caller
    """
    cfg = LoggingConfig()
    cfg.level = level
    cfg.progress_backend = "notebook"
    cfg.apply()


class LoggingConfig:  # pragma: nocover
    """
    Configuration for LensKit logging.

    This class is intended as a convenience for LensKit applications to set up a
    useful logging and progress reporting configuration; if unconfigured,
    LensKit will emit its logging messages directly to :mod:`structlog` and/or
    :mod:`logging`, which you can configure in any way you wish.

    Stability:
        Caller
    """

    level: int = logging.INFO
    stream: Literal["full", "simple", "json"] = "full"
    progress_backend: Literal["notebook", "rich"] | None = None
    file: Path | None = None
    file_level: int | None = None
    file_format: LogFormat = "json"

    def __init__(self):
        # initialize configuration from environment variables
        if ev_level := _env_level("LK_LOG_LEVEL"):
            self.level = ev_level

        if ev_file := os.environ.get("LK_LOG_FILE", None):
            self.file = Path(ev_file)

        if ev_level := _env_level("LK_LOG_FILE_LEVEL"):
            self.file_level = ev_level

    @property
    def effective_level(self) -> int:
        if self.file_level is not None and self.file_level < self.level:
            return self.file_level
        else:
            return self.level

    def set_stream_mode(self, mode: Literal["full", "simple", "json"]):
        """
        Configure the standard error stream mode.
        """
        self.stream = mode
        if mode == "full":
            self.force_console = True

    def set_verbose(self, verbose: bool | int = True):
        """
        Enable verbose logging.

        .. note::

            It is better to only call this method if your application's
            ``verbose`` option is provided, rather than passing your verbose
            option to it, to allow the ``LK_LOG_LEVEL`` environment variable to
            apply in the absence of a configuration option.

        Args:
            verbose:
                The level of verbosity.  Values of ``True`` or ``1`` turn on
                ``DEBUG``-level logs, and ``2`` or greater turns on tracing.
        """
        if isinstance(verbose, int) and verbose > 1:
            self.level = LVL_TRACE
        if verbose:
            self.level = logging.DEBUG
        else:
            self.level = logging.INFO

    def set_log_file(
        self, path: os.PathLike[str], level: int | None = None, format: LogFormat = "json"
    ):
        """
        Configure a log file.
        """
        self.file = Path(path)
        self.file_level = level
        self.file_format = format

    def apply(self):
        """
        Apply the configuration.
        """
        global _active_config

        if self.stream == "full":
            setup_console()

        root = logging.getLogger()

        if self.stream == "json":
            term = logging.StreamHandler(sys.stderr)
            term.setLevel(self.level)
            proc_fmt = structlog.processors.JSONRenderer()
        elif console.is_jupyter:
            term = logging.StreamHandler(sys.stdout)
            term.setLevel(self.level)
            proc_fmt = structlog.dev.ConsoleRenderer(
                colors=self.stream == "full" and not console.no_color
            )
        else:
            term = ConsoleHandler()
            term.setLevel(self.level)
            proc_fmt = structlog.dev.ConsoleRenderer(
                colors=self.stream == "full" and term.supports_color
            )

        eff_lvl = self.effective_level
        structlog.configure(
            processors=CORE_PROCESSORS + [structlog.stdlib.ProcessorFormatter.wrap_for_formatter],
            wrapper_class=lenskit_filtering_logger(eff_lvl),
            logger_factory=structlog.stdlib.LoggerFactory(),
        )
        formatter = structlog.stdlib.ProcessorFormatter(
            processors=[
                remove_internal,
                format_timestamp,
                proc_fmt,
            ],
            foreign_pre_chain=CORE_PROCESSORS,
        )

        term.setFormatter(formatter)
        root.addHandler(term)

        if self.file:
            file_level = self.file_level if self.file_level is not None else self.level
            file = logging.FileHandler(self.file, mode="w")

            if self.file_format == "json":
                proc_fmt = structlog.processors.JSONRenderer()
            elif self.file_format == "logfmt":
                proc_fmt = structlog.processors.LogfmtRenderer(key_order=["event", "timestamp"])
            else:
                proc_fmt = structlog.processors.KeyValueRenderer(key_order=["event", "timestamp"])

            ffmt = structlog.stdlib.ProcessorFormatter(
                processors=[
                    remove_internal,
                    structlog.processors.ExceptionPrettyPrinter(),
                    proc_fmt,
                ],
                foreign_pre_chain=CORE_PROCESSORS,
            )
            file.setFormatter(ffmt)
            file.setLevel(file_level)
            root.addHandler(file)

        root.setLevel(self.effective_level)

        if self.progress_backend is not None:
            set_progress_impl(self.progress_backend)
        elif self.stream == "full":
            set_progress_impl("rich")

        warnings.showwarning = log_warning
        warnings.filterwarnings("ignore", message=r"Sparse CSR tensor support is in beta state")

        _active_config = self


def _env_level(name: str) -> int | None:
    ev_level = os.environ.get(name, None)
    if ev_level:
        ev_level = ev_level.strip().upper()
        lmap = logging.getLevelNamesMapping()
        if re.match(r"^\d+$", ev_level):
            return int(ev_level)
        elif ev_level == "TRACE":
            return LVL_TRACE
        elif ev_level in lmap:
            return lmap[ev_level]
        else:
            warnings.warn(f"{name} set to invalid value {ev_level}")



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
LensKit test markers.
"""

import os

import pytest

jit_enabled = True
if "NUMBA_DISABLE_JIT" in os.environ:
    jit_enabled = False
if os.environ.get("PYTORCH_JIT", None) == "0":
    jit_enabled = False

wantjit = pytest.mark.skipif(not jit_enabled, reason="JIT required")



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Warning and error classes and general LensKit diagnostic code.
"""


class DataWarning(UserWarning):
    """
    Warning raised for detectable problems with input data.
    """

    pass


class DataError(Exception):
    """
    Error raised for detectable problesms with input data.
    """

    pass


class FieldError(KeyError):
    """
    The requested field does not exist.
    """

    def __init__(self, entity, field):
        super().__init__(f"{entity}[{field}]")


class ConfigWarning(UserWarning):
    """
    Warning raised for detectable problems with component configurations.
    """

    pass


class PipelineError(Exception):
    """
    Pipeline configuration errors.

    .. note::

        This exception is only to note problems with the pipeline configuration
        and structure (e.g. circular dependencies).  Errors *running* the
        pipeline are raised as-is.
    """


class PipelineWarning(Warning):
    """
    Pipeline configuration and setup warnings.  We also emit warnings to the
    logger in many cases, but this allows critical ones to be visible even if
    the client code has not enabled logging.

    .. note::

        This warning is only to note problems with the pipeline configuration
        and structure (e.g. circular dependencies).  Errors *running* the
        pipeline are raised as-is.
    """



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Support for computing chunks for parallel computation.
"""

# some defaults for chunking behavior
import math
from typing import NamedTuple

MIN_CHUNKABLE = 100
TGT_CHUNK_LIMIT = 1000
MIN_CHUNK_SIZE = 100
MAX_CHUNK_SIZE = 4000


class WorkChunks(NamedTuple):
    """
    The chunking configuration for parallel work.
    """

    total: int
    chunk_size: int
    chunk_count: int

    @classmethod
    def create(cls, njobs):
        """
        Compute the chunk size for parallel model training.
        """
        if njobs < MIN_CHUNKABLE:
            csize = njobs
            count = 1
        else:
            csize = max(njobs // TGT_CHUNK_LIMIT, MIN_CHUNK_SIZE)
            if csize > MAX_CHUNK_SIZE:
                csize = MAX_CHUNK_SIZE

            count = int(math.ceil(njobs / csize))

        return cls(njobs, csize, count)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Sparse matrix utility functions.
"""

# pyright: basic
from __future__ import annotations

import logging
from typing import Literal, overload

import scipy.sparse as sps
import torch

_log = logging.getLogger(__name__)


@overload
def normalize_sparse_rows(
    matrix: torch.Tensor, method: Literal["center"], inplace: bool = False
) -> tuple[torch.Tensor, torch.Tensor]: ...
@overload
def normalize_sparse_rows(
    matrix: torch.Tensor, method: Literal["unit"], inplace: bool = False
) -> tuple[torch.Tensor, torch.Tensor]: ...
def normalize_sparse_rows(
    matrix: torch.Tensor, method: str, inplace: bool = False
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Normalize the rows of a sparse matrix.
    """
    match method:
        case "unit":
            return _nsr_unit(matrix)
        case "center":
            return _nsr_mean_center(matrix)
        case _:
            raise ValueError(f"unsupported normalization method {method}")


def _nsr_mean_center(matrix: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
    nr, _nc = matrix.shape
    sums = matrix.sum(dim=1, keepdim=True).to_dense().reshape(nr)
    counts = torch.diff(matrix.crow_indices())
    assert sums.shape == counts.shape
    means = torch.nan_to_num(sums / counts, 0)
    return torch.sparse_csr_tensor(
        crow_indices=matrix.crow_indices(),
        col_indices=matrix.col_indices(),
        values=matrix.values() - torch.repeat_interleave(means, counts),
        size=matrix.shape,
    ), means


def _nsr_unit(matrix: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
    sqmat = torch.sparse_csr_tensor(
        crow_indices=matrix.crow_indices(),
        col_indices=matrix.col_indices(),
        values=matrix.values().square(),
    )
    norms = sqmat.sum(dim=1, keepdim=True).to_dense().reshape(matrix.shape[0])
    norms.sqrt_()
    recip_norms = torch.where(norms > 0, torch.reciprocal(norms), 0.0)
    return torch.sparse_csr_tensor(
        crow_indices=matrix.crow_indices(),
        col_indices=matrix.col_indices(),
        values=matrix.values() * torch.repeat_interleave(recip_norms, matrix.crow_indices().diff()),
        size=matrix.shape,
    ), norms


def torch_sparse_from_scipy(
    M: sps.coo_array, layout: Literal["csr", "coo", "csc"] = "coo"
) -> torch.Tensor:
    """
    Convert a SciPy :class:`sps.coo_array` into a torch sparse tensor.
    """
    ris = torch.from_numpy(M.row)
    cis = torch.from_numpy(M.col)
    vs = torch.from_numpy(M.data)
    indices = torch.stack([ris, cis])
    assert indices.shape == (2, M.nnz)
    T = torch.sparse_coo_tensor(indices, vs, size=M.shape)
    assert T.shape == M.shape

    match layout:
        case "csr":
            return T.to_sparse_csr()
        case "csc":
            return T.to_sparse_csc()
        case "coo":
            return T.coalesce()
        case _:
            raise ValueError(f"invalid layout {layout}")


def torch_sparse_to_scipy(M: torch.Tensor) -> sps.csr_array | sps.coo_array:
    if M.is_sparse:
        return sps.coo_array((M.values().numpy(), M.indices().numpy()), shape=M.shape)
    elif M.is_sparse_csr:
        return sps.csr_array(
            (M.values().numpy(), M.col_indices().numpy(), M.crow_indices().numpy()), shape=M.shape
        )
    else:
        raise TypeError("unsupported tensor type")



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Basic set statistics.
"""

from lenskit.data.items import ItemList

from ._base import Metric


class ListLength(Metric):
    """
    Report the length of the output (recommendation list or predictions).

    Stability:
        Caller
    """

    label = "N"  # type: ignore

    def __call__(self, recs: ItemList, test: ItemList) -> float:
        return len(recs)


class TestItemCount(Metric):
    """
    Report the number of test items.

    Stability:
        Caller
    """

    label = "TestItemCount"  # type: ignore

    def __call__(self, recs: ItemList, test: ItemList) -> float:
        return len(test)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
import warnings
from dataclasses import dataclass
from typing import Callable, TypeVar

import numpy as np
import pandas as pd

from lenskit.data import ItemList, ItemListCollection
from lenskit.diagnostics import DataWarning
from lenskit.logging import item_progress

from ._base import DecomposedMetric, GlobalMetric, ListMetric, Metric, MetricFunction

_log = logging.getLogger(__name__)
K1 = TypeVar("K1", bound=tuple)
K2 = TypeVar("K2", bound=tuple)


@dataclass(frozen=True)
class MetricWrapper:
    """
    Internal class for storing metrics.

    Stability:
        Internal
    """

    metric: Metric | MetricFunction
    label: str
    default: float | None = None

    @property
    def is_listwise(self) -> bool:
        "Check if this metric is listwise."
        return isinstance(self.metric, (ListMetric, Callable))

    @property
    def is_global(self) -> bool:
        "Check if this metric is global."
        return isinstance(self.metric, GlobalMetric)

    @property
    def is_decomposed(self) -> bool:
        "Check if this metric is decomposed."
        return isinstance(self.metric, DecomposedMetric)

    def measure_list(self, list: ItemList, test: ItemList) -> float:
        if isinstance(self.metric, ListMetric):
            return self.metric.measure_list(list, test)
        elif isinstance(self.metric, Callable):
            return self.metric(list, test)
        else:
            raise TypeError(f"metric {self.metric} does not support list measurement")

    def measure_run(self, run: ItemListCollection, test: ItemListCollection) -> float:
        if isinstance(self.metric, GlobalMetric):
            return self.metric.measure_run(run, test)
        else:
            raise TypeError(f"metric {self.metric} does not support global measurement")


class RunAnalysisResult:
    """
    Results of a bulk metric computation.

    Stability:
        Caller
    """

    _list_metrics: pd.DataFrame
    _global_metrics: pd.Series
    _defaults: dict[str, float]

    def __init__(self, lmvs: pd.DataFrame, gmvs: pd.Series, defaults: dict[str, float]):
        self._list_metrics = lmvs
        self._global_metrics = gmvs
        self._defaults = defaults

    def global_metrics(self) -> pd.Series:
        """
        Get the global metric scores.  This is only the results of
        global metrics; it does not include aggregates of per-list metrics.  For
        aggregates of per-list metrics, call :meth:`list_summary`.
        """
        return self._global_metrics

    def list_metrics(self, fill_missing=True) -> pd.DataFrame:
        """
        Get the per-list scores of the results.  This is a data frame with one
        row per list (with the list key on the inded), and one metric per
        column.

        Args:
            fill_missing:
                If ``True`` (the default), fills in missing values with each
                metric's default value when available.  Pass ``False`` if you
                want to do analyses that need to treat missing values
                differently.
        """
        return self._list_metrics.fillna(self._defaults)

    def list_summary(self, *keys: str) -> pd.DataFrame:
        """
        Sumamry statistics for the per-list metrics.  Each metric is on its own
        row, with columns reporting the following:

        ``mean``:
            The mean metric value.
        ``median``:
            The median metric value.
        ``std``:
            The (sample) standard deviation of the metric.

        Additional columns are added based on other options.  Missing metric
        values are filled with their defaults before computing statistics.

        Args:
            keys:
                Identifiers for different conditions that should be reported
                separately (grouping keys for the final result).
        """
        scores = self.list_metrics(fill_missing=True)
        if keys:
            df = scores.groupby(list(keys)).agg(["mean", "median", "std"]).stack(level=0)
            assert isinstance(df, pd.DataFrame)
        else:
            df = scores.agg(["mean", "median", "std"]).T
            df.index.name = "metric"
        return df

    def merge_from(self, other: RunAnalysisResult):
        """
        Merge another set of analysis results into this one.
        """
        for c in self._list_metrics.columns:
            if c in other._list_metrics.columns:
                warnings.warn(f"list metric {c} appears in both merged results", DataWarning)

        for c in self._global_metrics.index:
            if c in other._global_metrics.index:
                warnings.warn(f"global metric {c} appears in both merged results", DataWarning)

        self._list_metrics = self._list_metrics.join(other._list_metrics, how="outer")
        self._global_metrics = pd.concat([self._global_metrics, other._global_metrics])


def _wrap_metric(
    m: Metric | MetricFunction | type[Metric],
    label: str | None = None,
    default: float | None = None,
) -> MetricWrapper:
    if isinstance(m, type):
        m = m()

    if label is None:
        if isinstance(m, Metric):
            wl = m.label
        else:
            wl = m.__name__  # type: ignore
    else:
        wl = label

    if default is None:
        if isinstance(m, ListMetric):
            default = m.default
        else:
            default = 0.0

        if default is not None and not isinstance(default, (float, int, np.floating, np.integer)):
            raise TypeError(f"metric {m} has unsupported default {default}")

    return MetricWrapper(m, wl, default)  # type: ignore


class RunAnalysis:
    """
    Compute metrics over a collection of item lists composing a run.

    Args:
        metrics:
            A list of metrics; you can also add them with :meth:`add_metric`,
            which provides more flexibility.

    Stability:
        Caller
    """

    metrics: list[MetricWrapper]
    "The list of metrics to compute."

    def __init__(self, *metrics: Metric):
        self.metrics = [_wrap_metric(m) for m in metrics]

    def add_metric(
        self,
        metric: Metric | MetricFunction | type[Metric],
        label: str | None = None,
        default: float | None = None,
    ):
        """
        Add a metric to this metric set.

        Args:
            metric:
                The metric to add to the set.
            label:
                The label to use for the metric's results.  If unset, obtains
                from the metric.
            default:
                The default value to use in aggregates when a user does not have
                recommendations. If unset, obtains from the metric's ``default``
                attribute (if specified), or 0.0.
        """
        self.metrics.append(_wrap_metric(metric, label, default))

    def compute(
        self, outputs: ItemListCollection[K1], test: ItemListCollection[K2]
    ) -> RunAnalysisResult:
        """
        Deprecated alias for :meth:`measure`.

        .. deprecated:: 2025.1.1
            Use :meth:`measure` instead.
        """
        return self.measure(outputs, test)

    def measure(
        self, outputs: ItemListCollection[K1], test: ItemListCollection[K2]
    ) -> RunAnalysisResult:
        """
        Measure a set of outputs against a set of dest data.
        """
        self._validate_setup()
        index = pd.MultiIndex.from_tuples(outputs.keys())
        index.names = list(outputs.key_fields)

        lms = [m for m in self.metrics if m.is_listwise or m.is_decomposed]
        gms = [m for m in self.metrics if m.is_global]
        list_results = pd.DataFrame({m.label: np.nan for m in lms}, index=index)
        list_intermediates = {m.label: [] for m in self.metrics if m.is_decomposed}

        n = len(outputs)
        _log.info("computing %d listwise metrics for %d output lists", len(lms), n)
        with item_progress("Measuring", n) as pb:
            for i, (key, out) in enumerate(outputs):
                list_test = test.lookup_projected(key)
                if out is None:
                    pass
                elif list_test is None:
                    _log.warning("list %s: no test items", key)
                else:
                    l_row: list[float | None] = []
                    for m in lms:
                        mv = None
                        if m.is_decomposed:
                            assert isinstance(m.metric, DecomposedMetric)
                            val = m.metric.compute_list_data(out, list_test)
                            list_intermediates[m.label].append(val)
                            mv = m.metric.extract_list_metric(val)
                        if mv is None and m.is_listwise:
                            mv = m.measure_list(out, list_test)
                        l_row.append(mv)
                    list_results.iloc[i] = l_row  # type: ignore
                pb.update()

        _log.info("computing %d global metrics for %d output lists", len(gms), n)
        global_results = pd.Series(
            {
                m.label: (
                    m.metric.global_aggregate(list_intermediates[m.label])
                    if m.is_decomposed
                    else m.measure_run(outputs, test)
                )
                for m in self.metrics
                if m.is_global or m.is_decomposed
            },
            dtype=np.float64,
        )

        return RunAnalysisResult(
            list_results,
            global_results,
            {m.label: m.default for m in self.metrics if m.default is not None},
        )

    def _validate_setup(self):
        seen = set()
        for m in self.metrics:
            lbl = m.label
            if lbl in seen:
                raise RuntimeError(f"duplicate metric: {lbl}")
            seen.add(lbl)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: strict
import logging
from typing import Generic, Iterable, Iterator

from .invoker import A, InvokeOp, M, ModelOpInvoker, R

_log = logging.getLogger(__name__)


class InProcessOpInvoker(ModelOpInvoker[A, R], Generic[M, A, R]):
    model: M
    function: InvokeOp[M, A, R]

    def __init__(self, model: M, func: InvokeOp[M, A, R]):
        _log.debug("setting up in-process worker")
        self.model = model
        self.function = func

    def map(self, tasks: Iterable[A]) -> Iterator[R]:
        for task in tasks:
            res = self.function(self.model, task)
            yield res

    def shutdown(self):
        del self.model



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Metrics for evaluating recommender outputs.
"""

from typing import Callable, ParamSpec

from lenskit.data import ItemList

from ._base import GlobalMetric, ListMetric, Metric, MetricFunction
from ._quick import quick_measure_model
from .basic import ListLength, TestItemCount
from .bulk import RunAnalysis, RunAnalysisResult
from .predict import MAE, RMSE
from .ranking import NDCG, RBP, Hit, MeanPopRank, Precision, RankingMetricBase, Recall, RecipRank

__all__ = [
    "Metric",
    "MetricFunction",
    "ListMetric",
    "GlobalMetric",
    "RankingMetricBase",
    "RunAnalysis",
    "RunAnalysisResult",
    "ListLength",
    "TestItemCount",
    "RMSE",
    "MAE",
    "NDCG",
    "RBP",
    "Hit",
    "Precision",
    "Recall",
    "RecipRank",
    "MeanPopRank",
    "quick_measure_model",
]

P = ParamSpec("P")


def call_metric(
    metric: ListMetric | MetricFunction | Callable[P, ListMetric],
    outs: ItemList,
    test: ItemList | None = None,
    *args: P.args,
    **kwargs: P.kwargs,
) -> float:
    """
    Call a metric, instantiating it if necessary.  This intended to be a quick
    convenience when you just need to call a metric with its default settings
    and don't want to mess around with object/class distinctions.  You usually
    don't actually want to use it.

    Supports both the base :class:`Metric` protocol and the extensions in
    :class:`PredictMetric`.

    Args:
        metric:
            The metric to call.  Note that the type of ``Callable`` is slightly
            too loose, it actually needs to be a class that subclasses
            :class:`Metric`.
        outs:
            The output to measure.
        test:
            The test data to measure.
    """

    if isinstance(metric, type):
        metric = metric(*args, **kwargs)

    if isinstance(metric, ListMetric):
        return metric.measure_list(outs, test)  # type: ignore
    elif isinstance(metric, Callable):
        return metric(outs, test)  # type: ignore
    else:  # pragma: nocover
        raise TypeError(f"invalid metric {metric}")



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
LensKit ALS implementations.
"""

from ._common import ALSBase, ALSConfig
from ._explicit import BiasedMFConfig, BiasedMFScorer
from ._implicit import ImplicitMFConfig, ImplicitMFScorer

__all__ = [
    "ALSBase",
    "ALSConfig",
    "BiasedMFScorer",
    "BiasedMFConfig",
    "ImplicitMFScorer",
    "ImplicitMFConfig",
]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Entity accessors for data sets.
"""

from __future__ import annotations

import numpy as np
import pandas as pd
import pyarrow as pa
from typing_extensions import Any, overload

from lenskit.logging import get_logger

from .attributes import AttributeSet, attr_set
from .schema import EntitySchema
from .types import IDArray, IDSequence
from .vocab import Vocabulary

_log = get_logger(__name__)


class EntitySet:
    """
    Representation of a set of entities from the dataset.  Obtained from
    :meth:`Dataset.entities`.

    Client code does not need to construct this class; obtain instances from a
    dataset's :meth:`~lenskit.data.Dataset.entities` method.
    """

    name: str
    """
    The name of the entity class for these entities.
    """
    schema: EntitySchema
    vocabulary: Vocabulary
    """
    The identifier vocabulary for this schema.
    """
    _table: pa.Table
    """
    The Arrow table of entity information.
    """
    _selected: pa.Int32Array | None = None
    """
    The indices of the selected entities.
    """

    def __init__(
        self,
        name: str,
        schema: EntitySchema,
        vocabulary: Vocabulary,
        table: pa.Table,
        _sel: pa.Int32Array | None = None,
    ):
        self.name = name
        self.schema = schema
        self.vocabulary = vocabulary
        self._table = table
        self._selected = _sel

    @property
    def attributes(self) -> list[str]:
        """
        Get the attribute names for this enrtity class.
        """
        return list(self.schema.attributes.keys())

    def count(self) -> int:
        """
        Return the number of entities in this entity set.
        """
        if self._selected is not None:
            return len(self._selected)
        else:
            return self._table.num_rows

    def ids(self) -> IDArray:
        """
        Get the identifiers of the entities in this set.  This is returned
        directly as PyArrow array instead of NumPy.
        """
        if self._selected is not None:
            return self.vocabulary.ids(self._selected.to_numpy())
        else:
            return self.vocabulary.ids()

    def numbers(self) -> np.ndarray[int, np.dtype[np.int32]]:
        """
        Get the numbers (from the vocabulary) for the entities in this set.
        """
        if self._selected is not None:
            return self._selected.to_numpy()
        else:
            return np.arange(self.count(), dtype=np.int32)

    def arrow(self) -> pa.Table:
        """
        Get these entities and their attributes as a PyArrow table.
        """
        if self._selected is not None:
            return self._table.take(self._selected)
        else:
            return self._table

    def pandas(self) -> pd.DataFrame:
        """
        Get the entities and their attributes as a Pandas data frame.
        """
        return self.arrow().to_pandas()

    def attribute(self, name: str) -> AttributeSet:
        """
        Get values of an attribute for the entites in this entity set.
        """
        spec = self.schema.attributes[name]

        return attr_set(name, spec, self._table, self.vocabulary, self._selected)

    @overload
    def select(self, *, ids: IDSequence | None = None) -> EntitySet: ...
    @overload
    def select(
        self,
        *,
        numbers: np.ndarray[int, np.dtype[np.integer[Any]]] | pa.IntegerArray[Any] | None = None,
    ) -> EntitySet: ...
    def select(
        self,
        *,
        ids: IDSequence | None = None,
        numbers: np.ndarray[int, np.dtype[np.integer[Any]]] | pa.IntegerArray[Any] | None = None,
    ) -> EntitySet:
        """
        Select a subset of the entities in this set.

        .. note::

            The vocabulary is unchanged, so numbers in the resulting set will be
            entity numbers in the dataset's vocabulary.  They are not rearranged
            to be relative to this entity set.

        Args:
            ids:
                The entity identifiers to select.
            numbers:
                The entity numbers to select.

        Returns:
            The entity subset.
        """
        if numbers is not None and ids is None:
            picked = pa.array(numbers).cast(pa.int32())
        elif ids is not None and numbers is None:
            picked = pa.array(self.vocabulary.numbers(ids)).cast(pa.int32())
        else:  # pragma: nocover
            raise ValueError("specify exactly one of ids and numbers")

        return EntitySet(self.name, self.schema, self.vocabulary, self._table, picked)  # type: ignore

    def __len__(self):
        return self.count()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Relationship accessors for Dataset.
"""

from __future__ import annotations

import warnings
from collections.abc import Mapping
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import scipy.sparse as sps
import torch
from numpy.typing import NDArray
from typing_extensions import Literal, overload, override

from lenskit.diagnostics import DataWarning, FieldError
from lenskit.logging import get_logger
from lenskit.random import random_generator

from .items import ItemList
from .matrix import COOStructure, CSRStructure
from .schema import RelationshipSchema, id_col_name, num_col_name
from .types import ID, LAYOUT, MAT_AGG
from .vocab import Vocabulary

if TYPE_CHECKING:
    from .dataset import Dataset

_log = get_logger(__name__)


class RelationshipSet:
    """
    Representation for a set of relationship records.  This is the class for
    accessing general relationships, with arbitrarily many entity classes
    involved and repeated relationships allowed.

    For two-entity relationships without duplicates (including relationships
    formed by coalescing repeated relationships or interactions),
    :class:`MatrixRelationshipSet` extends this with additional capabilities.

    Relationship sets can be pickled or serialized, and will not save the entire
    dataset with them.  They are therefore safe to save as component elements
    during training processes.

    .. note::

        Client code does not need to construct this class; obtain instances from
        a dataset's :meth:`~lenskit.data.Dataset.relationships` or
        :meth:`~lenskit.data.Dataset.interactions` method.
    """

    name: str
    """
    The name of the relationship class for these relationships.
    """
    schema: RelationshipSchema

    _table: pa.Table
    """
    The Arrow table of relationship information.
    """

    _vocabularies: dict[str, Vocabulary]
    _link_cols: list[str]

    def __init__(
        self,
        ds: Dataset,
        name: str,
        schema: RelationshipSchema,
        table: pa.Table,
    ):
        self.name = name
        self.schema = schema
        self._table = table

        self._vocabularies = {e: ds.entities(e).vocabulary for e in schema.entities}
        self._link_cols = [num_col_name(e) for e in schema.entities]

    @property
    def is_interaction(self) -> bool:
        """
        Query whether these relationships represent interactions.
        """
        return self.schema.interaction

    @property
    def attribute_names(self) -> list[str]:
        return [c for c in self._table.column_names if c not in self._link_cols]

    def count(self):
        if "count" in self._table.column_names:  # pragma: nocover
            raise NotImplementedError()

        return self._table.num_rows

    def arrow(self, *, attributes: str | list[str] | None = None, ids=False) -> pa.Table:
        """
        Get these relationships and their attributes as a PyArrow table.

        Args:
            attributes:
                The attributes to select.
            ids:
                If ``True``, include ID columns for the entities, instead of
                just the number columns.
        """
        table = self._table
        cols = self._link_cols

        if ids:
            id_cols = {}
            for e in self.schema.entity_class_names:
                id_cols[id_col_name(e)] = pa.array(
                    self._vocabularies[e].ids(table.column(num_col_name(e)).to_numpy())
                )
            id_tbl = pa.table(id_cols)
            cols = id_tbl.column_names
            for col in table.column_names:
                if col in self._link_cols:
                    continue

                id_tbl = id_tbl.append_column(col, table.column(col))
            table = id_tbl

        if attributes is not None:
            if isinstance(attributes, str):
                attr_cols = [attributes]
            else:
                attr_cols = attributes
            for ac in attr_cols:
                if ac not in table.column_names:
                    raise FieldError(self.name, ac)
            table = table.select(cols + attr_cols)

        return table

    def pandas(self, *, attributes: str | list[str] | None = None, ids=False) -> pd.DataFrame:
        """
        Get these relationship and their attributes as a PyArrow table.

        Args:
            attributes:
                The attributes to include in the resulting table.
            ids:
                If ``True``, include ID columns for the entities, instead of
                just the number columns.
        """
        tbl = self.arrow(attributes=attributes, ids=ids)
        return tbl.to_pandas()

    def matrix(
        self, *, combine: MAT_AGG | Mapping[str, MAT_AGG] | None = None
    ) -> MatrixRelationshipSet:  # pragma: nocover
        """
        Convert this relationship set into a matrix, coalescing duplicate
        observations.

        Args:
            combine:
                The method for combining attribute values for repeated
                relationships or interactions.  Can either be a single strategy
                or a mapping from attribute names to combination strategies.
        """
        raise NotImplementedError()


class MatrixRelationshipSet(RelationshipSet):
    """
    Two-entity relationships without duplicates, accessible in matrix form.

    .. note::

        Client code does not need to construct this class; obtain instances from
        a relationship set's :meth:`~RelationshipSet.matrix` method.
    """

    _row_ptrs: np.ndarray[int, np.dtype[np.int32]]
    row_vocabulary: Vocabulary
    row_type: str
    _row_stats: pd.DataFrame | None = None

    col_vocabulary: Vocabulary
    col_type: str
    _col_stats: pd.DataFrame | None = None

    rc_index: pd.Index

    def __init__(
        self,
        ds: Dataset,
        name: str,
        schema: RelationshipSchema,
        table: pa.Table,
    ):
        super().__init__(ds, name, schema, table)
        # order the table to compute the sparse matrix
        entities = list(schema.entities.keys())
        row, col = entities
        row_col_name = num_col_name(row)
        col_col_name = num_col_name(col)

        self.row_type = row
        self.row_vocabulary = ds.entities(row).vocabulary
        self.col_type = col
        self.col_vocabulary = ds.entities(col).vocabulary

        e_cols = [num_col_name(e) for e in entities]
        table = table.sort_by([(c, "ascending") for c in e_cols])

        # compute the row pointers
        n_rows = len(self.row_vocabulary)
        row_sizes = np.zeros(n_rows + 1, dtype=np.int32())
        rsz_struct = pc.value_counts(table.column(e_cols[0]))
        rsz_nums = rsz_struct.field("values")
        rsz_counts = rsz_struct.field("counts").cast(pa.int32())
        row_sizes[np.asarray(rsz_nums) + 1] = rsz_counts
        self._row_ptrs = np.cumsum(row_sizes, dtype=np.int32)
        self._table = table

        # make the index
        self.rc_index = pd.Index(
            self._rc_combined_nums(
                self._table.column(row_col_name).to_numpy(),
                self._table.column(col_col_name).to_numpy(),
            )
        )

    @property
    def n_rows(self):
        return len(self.row_vocabulary)

    @property
    def n_cols(self) -> int:
        return len(self.col_vocabulary)

    @override
    def matrix(
        self, *, combine: MAT_AGG | dict[str, MAT_AGG] | None = None
    ) -> MatrixRelationshipSet:
        # already a matrix relationship set
        return self

    def csr_structure(self) -> CSRStructure:
        """
        Get the compressed sparse row structure of this relationship matrix.
        """
        n_rows = len(self.row_vocabulary)
        n_cols = len(self.col_vocabulary)

        colinds = self._table.column(num_col_name(self.col_type)).to_numpy()
        return CSRStructure(self._row_ptrs, colinds, (n_rows, n_cols))

    def coo_structure(self) -> COOStructure:
        """
        Get the compressed sparse row structure of this relationship matrix.
        """
        n_rows = len(self.row_vocabulary)
        n_cols = len(self.col_vocabulary)

        rowinds = self._table.column(num_col_name(self.row_type)).to_numpy()
        colinds = self._table.column(num_col_name(self.col_type)).to_numpy()
        return COOStructure(rowinds, colinds, (n_rows, n_cols))

    @overload
    def scipy(
        self, attribute: str | None = None, *, layout: Literal["coo"], legacy: Literal[True]
    ) -> sps.coo_matrix: ...
    @overload
    def scipy(
        self,
        attribute: str | None = None,
        *,
        layout: Literal["coo"],
        legacy: Literal[False] = False,
    ) -> sps.coo_array: ...
    @overload
    def scipy(
        self, attribute: str | None = None, *, layout: Literal["csr"] = "csr", legacy: Literal[True]
    ) -> sps.csr_matrix: ...
    @overload
    def scipy(
        self,
        attribute: str | None = None,
        *,
        layout: Literal["csr"] = "csr",
        legacy: Literal[False] = False,
    ) -> sps.csr_array: ...
    @overload
    def scipy(
        self, attribute: str | None = None, *, layout: LAYOUT = "csr", legacy: bool = False
    ) -> sps.sparray | sps.spmatrix: ...
    def scipy(
        self, attribute: str | None = None, *, layout: LAYOUT = "csr", legacy: bool = False
    ) -> sps.sparray | sps.spmatrix:
        """
        Get this relationship matrix as a SciPy sparse matrix.

        Args:
            attribute:
                The attribute to return, or ``None`` to return an indicator-only
                sparse matrix (all observed values are 1).
            layout:
                The matrix layout to return.

        Returns:
            The sparse matrix.
        """
        n_rows = self.n_rows
        n_cols = self.n_cols
        nnz = self._table.num_rows

        colinds = self._table.column(num_col_name(self.col_type)).to_numpy()
        if attribute is None or (attribute == "count" and "count" not in self._table.column_names):
            values = np.ones(nnz, dtype=np.float32)
        else:
            values = self._table.column(attribute).to_numpy()

        if layout == "csr":
            if legacy:
                return sps.csr_matrix((values, colinds, self._row_ptrs), shape=(n_rows, n_cols))
            else:
                return sps.csr_array((values, colinds, self._row_ptrs), shape=(n_rows, n_cols))
        elif layout == "coo":
            rowinds = self._table.column(num_col_name(self.row_type))
            if legacy:
                return sps.coo_matrix((values, (rowinds, colinds)), shape=(n_rows, n_cols))
            else:
                return sps.coo_array((values, (rowinds, colinds)), shape=(n_rows, n_cols))

    @overload
    def torch(
        self, attribute: str | None = None, *, layout: Literal["csr"] = "csr"
    ) -> torch.Tensor: ...
    @overload
    def torch(self, attribute: str | None = None, *, layout: Literal["coo"]) -> torch.Tensor: ...
    def torch(self, attribute: str | None = None, *, layout: LAYOUT = "csr") -> torch.Tensor:
        """
        Get this relationship matrix as a PyTorch sparse tensor.

        Args:
            attribute:
                The attribute to return, or ``None`` to return an indicator-only
                sparse matrix (all observed values are 1).
            layout:
                The matrix layout to return.

        Returns:
            The sparse matrix.
        """
        n_rows = self.n_rows
        n_cols = self.n_cols
        nnz = self._table.num_rows

        colinds = self._table.column(num_col_name(self.col_type)).to_numpy()
        colinds = torch.tensor(colinds)
        if attribute is None or (attribute == "count" and "count" not in self._table.column_names):
            values = torch.ones(nnz, dtype=torch.float32)
        elif pa.types.is_timestamp(self._table.field(attribute).type):
            vals = self._table.column(attribute)
            vals = vals.cast(pa.timestamp("s")).cast(pa.int64())
            values = torch.tensor(vals.to_numpy())
        else:
            values = torch.tensor(self._table.column(attribute).to_numpy())

        if layout == "csr":
            return torch.sparse_csr_tensor(
                crow_indices=torch.tensor(self._row_ptrs),
                col_indices=colinds,
                values=values,
                size=(n_rows, n_cols),
            )
        elif layout == "coo":
            rowinds = torch.tensor(self._table.column(num_col_name(self.row_type)).to_numpy())
            indices = torch.stack((rowinds, colinds))
            return torch.sparse_coo_tensor(
                indices=indices, values=values, size=(n_rows, n_cols)
            ).coalesce()

    def sample_negatives(
        self,
        rows: np.ndarray[int, np.dtype[np.int32]],
        *,
        weighting: Literal["uniform", "popularity"] = "uniform",
        n: int | None = None,
        verify: bool = True,
        max_attempts: int = 10,
        rng: np.random.Generator | None = None,
    ) -> NDArray[np.int32]:
        """
        Sample negative columns (columns with no observation recorded) for an
        array of rows. On a normal interaction matrix, this samples negative
        items for users.

        Args:
            rows:
                The row numbers.  Duplicates are allowed, and negative columns
                are sampled independently for each row. Must be a 1D array or
                tensor.
            weighting:
                The weighting for sampled negatives; ``uniform`` samples them
                uniformly at random, while ``popularity`` samples them
                proportional to their popularity (number of occurrences).
            n:
                The number of negatives to sample for each user.  If ``None``,
                a single-dimensional vector is returned.
            verify:
                Whether to verify that the negative items are actually negative.
                Unverified sampling is much faster but can return false
                negatives.
            max_attempts:
                When verification is on, the maximum attempts before giving up
                and returning a possible false negative.
            rng:
                A random number generator to use.
        """
        rng = random_generator(rng)

        _log.debug("samping negatives", nrows=len(rows), ncols=n)
        if n is None:
            shape = len(rows)
        else:
            shape = (len(rows), n)

        match weighting:
            case "uniform":
                columns = rng.choice(self.n_cols, size=shape, replace=True)
            case "popularity":
                ccol = self._table.column(num_col_name(self.col_type)).to_numpy()
                trows = rng.choice(self._table.num_rows, size=shape, replace=True)
                columns = ccol[trows]
        columns = np.require(columns, "i4")

        if verify:
            if n is None:
                self._check_negatives_and_resample(rows, columns, max_attempts, rng, weighting)
            else:
                for c in range(n):
                    self._check_negatives_and_resample(
                        rows, columns[:, c], max_attempts, rng, weighting
                    )

        return columns

    def _check_negatives(
        self, rows: NDArray[np.int32], columns: NDArray[np.int32]
    ) -> NDArray[np.bool]:
        nums = self._rc_combined_nums(rows, columns)
        locs = self.rc_index.get_indexer_for(nums)
        return locs >= 0

    def _rc_combined_nums(self, rows: NDArray[np.int32], columns: NDArray[np.int32]):
        rnums = rows.astype(np.uint64)
        cnums = columns.astype(np.uint64)
        return (rnums << 32) + cnums

    def _check_negatives_and_resample(
        self,
        rows: NDArray[np.int32],
        columns: NDArray[np.int32],
        max_attempts: int,
        rng: np.random.Generator,
        weighting: Literal["uniform", "popularity"],
    ):
        non_neg = self._check_negatives(rows, columns)
        _log.debug("checking negatives", nrows=len(rows), npos=np.sum(non_neg).item())
        if np.any(non_neg):
            if max_attempts > 0:
                columns[non_neg] = self.sample_negatives(
                    rows[non_neg],
                    verify=True,
                    rng=rng,
                    max_attempts=max_attempts - 1,
                    weighting=weighting,
                )
            else:
                warnings.warn(
                    "failed to find verified negatives for {} users".format(np.sum(non_neg)),
                    DataWarning,
                )

    def row_table(self, id: ID | None = None, *, number: int | None = None) -> pa.Table | None:
        """
        Get a single row of this interaction matrix as a table.
        """
        if number is None and id is None:  # pragma: noover
            raise ValueError("must provide one of id and number")

        if number is None:
            number = self.row_vocabulary.number(id, "none")
            if number is None:
                return None

        row_start = self._row_ptrs[number]
        row_end = self._row_ptrs[number + 1]

        tbl = self._table.slice(row_start, row_end - row_start)
        tbl = tbl.drop_columns(num_col_name(self.row_type))
        return tbl

    def row_items(self, id: ID | None = None, *, number: int | None = None) -> ItemList | None:
        """
        Get a single row of this interaction matrix as an item list.  Only valid
        when the column entity class is ``item''.
        """
        if self.col_type != "item":
            raise RuntimeError("row_items() only valid for item-column matrices")

        tbl = self.row_table(id=id, number=number)
        if tbl is None:
            return None

        return ItemList.from_arrow(tbl, vocabulary=self.col_vocabulary)

    def row_stats(self):
        if self._row_stats is None:
            self._row_stats = self._compute_stats(self.row_type, self.col_type, self.row_vocabulary)
        return self._row_stats

    def col_stats(self):
        if self._col_stats is None:
            self._col_stats = self._compute_stats(self.col_type, self.row_type, self.col_vocabulary)
        return self._col_stats

    def _compute_stats(
        self, stat_type: str, other_type: str, stat_vocab: Vocabulary
    ) -> pd.DataFrame:
        s_col = num_col_name(stat_type)
        group = self._table.group_by(s_col)
        o_col = num_col_name(other_type)
        aggs = [(o_col, "count"), (o_col, "count_distinct")]
        if "count" in self._table.column_names:
            aggs.append(("count", "sum"))
        if "rating" in self._table.column_names:
            aggs += [("rating", "count"), ("rating", "mean")]
        if "first_time" in self._table.column_names:
            aggs.append(("first_time", "min"))
        if "last_time" in self._table.column_names:
            aggs.append(("last_time", "max"))
        if "timestamp" in self._table.column_names:
            aggs += [("timestamp", "min"), ("timestamp", "max")]

        stats = group.aggregate(aggs).to_pandas()
        stats = stats.rename(
            columns={
                o_col + "_count": "record_count",
                o_col + "_count_distinct": other_type + "_count",
                "count_sum": "count",
                "rating_mean": "mean_rating",
                "first_time_min": "first_time",
                "last_time_max": "last_time",
            }
        )
        if "count" not in stats.columns:
            stats["count"] = stats["record_count"]
        if "timestamp" in self._table.column_names:
            if "first_time" in self._table.columns:
                stats["first_time"] = np.minimum(stats["first_time"], stats["timestamp_min"])
            else:
                stats["first_time"] = stats["timestamp_min"]
            del stats["timestamp_min"]

            if "last_time" in self._table.columns:
                stats["last_time"] = np.maximum(stats["last_time"], stats["timestamp_max"])
            else:
                stats["last_time"] = stats["timestamp_max"]
            del stats["timestamp_max"]

        id_col = id_col_name(stat_type)
        stats[id_col] = stat_vocab.ids(stats[s_col])
        del stats[s_col]
        stats.set_index(id_col, inplace=True)
        stats = stats.reindex(stat_vocab.index, fill_value=0)
        if "mean_rating" in stats.columns:
            stats.loc[stats["rating_count"] == 0, "mean_rating"] = np.nan
        if "first_time" in stats.columns:
            stats.loc[stats["count"] == 0, "first_time"] = pd.NaT
        if "last_time" in stats.columns:
            stats.loc[stats["count"] == 0, "last_time"] = pd.NaT

        return stats



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import click

from .convert import convert
from .describe import describe
from .fetch import fetch


@click.group
def data():
    """
    Data conversion and processing commands.
    """
    pass


data.add_command(convert)
data.add_command(describe)
data.add_command(fetch)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT
"""
Multi-format array layout.
"""

# pyright: basic
from __future__ import annotations

import numpy as np
import pyarrow as pa
import torch
from numpy.typing import ArrayLike, NDArray
from typing_extensions import Generic, Literal, LiteralString, Sequence, TypeVar, overload

NPT = TypeVar("NPT", bound=np.generic)


class MTArray(Generic[NPT]):
    """
    Multi-typed array class, allowing arrays to be easily converted between
    NumPy, PyTorch, and other supported backends, caching the conversion result.

    We use this class instead of one canonical format so that data can be
    converted lazily, and can be left in-place when passing from one component
    to another that use the same computational engine.

    .. note::

        This class is only intended for read-only arrays.  It is **not defined**
        whether the different arrays share storage, and modifying one format may
        or may not modify another.  For example, PyTorch and NumPy usually share
        storage when both are on CPU, but a GPU tensor and CPU ndarray do not.

    Stability:
        Internal
    """

    _shape: tuple[int, ...] | None = None
    _unknown: object = None
    _numpy: NDArray[NPT] | None = None
    _arrow: pa.Array | pa.Tensor | None = None
    _torch: torch.Tensor | None = None

    def __init__(
        self, array: NDArray[NPT] | torch.Tensor | pa.Array | pa.Tensor | Sequence | ArrayLike
    ):
        """
        Construct a new MTArray around an array.
        """
        # TODO: support DLpack
        if isinstance(array, torch.Tensor):
            # torch might not be on-device
            self._torch = array
            self._shape = array.shape
        else:
            # stash it in the common-format field for lazy conversion
            self._unknown = array
            # save the shape for well-known types
            if isinstance(array, pa.Array):
                self._shape = (len(array),)
            elif isinstance(array, (pa.Tensor, np.ndarray)):
                self._shape = array.shape

    @property
    def shape(self) -> tuple[int, ...]:
        if self._shape is None:
            self._shape = self.numpy().shape

        return self._shape

    def numpy(self) -> NDArray[NPT]:
        """
        Get the array as a NumPy array.
        """
        if self._numpy is None:
            self._numpy = np.asarray(self._convertible())

        assert self._numpy is not None
        return self._numpy

    def torch(self, *, device: str | None = None) -> torch.Tensor:
        """
        Get the array as a PyTorch tensor.

        Args:
            device:
                The device on which the Torch tensor should reside.
        """
        if self._torch is None:
            if torch.is_tensor(self._unknown):
                self._torch = self._unknown
            else:
                # Make sure we have a writeable array for Torch. Client code
                # still shouldn't write to it.
                arr = np.require(self.numpy(), requirements="W")
                return torch.tensor(arr)

        if device:
            return self._torch.to(device)
        else:
            return self._torch

    def arrow(self) -> pa.Array | pa.Tensor:
        """
        Get the array as an Arrow :class:`~pyarrow.Array`.
        """
        if self._arrow is None:
            if isinstance(self._unknown, (pa.Array, pa.Tensor)):
                self._arrow = self._unknown
            else:
                arr = self.numpy()
                if len(arr.shape) == 1:  # type: ignore
                    self._arrow = pa.array(arr)  # type: ignore
                else:
                    self._arrow = pa.Tensor.from_numpy(arr)  # type: ignore

        assert self._arrow is not None
        return self._arrow

    @overload
    def to(self, format: Literal["numpy"]) -> NDArray[NPT]: ...
    @overload
    def to(self, format: Literal["torch"], *, device: str | None) -> torch.Tensor: ...
    @overload
    def to(self, format: Literal["arrow"]) -> pa.Array | pa.Tensor: ...
    @overload
    def to(self, format: LiteralString, *, device: str | None = None) -> ArrayLike: ...
    def to(self, format: str, *, device: str | None = None) -> object:
        """
        Obtain the array in the specified format (dynamic version).
        """
        match format:
            case "numpy":
                return self.numpy()
            case "torch":
                return self.torch(device=device)
            case "arrow":
                return self.arrow()
            case _:
                raise RuntimeError(f"unsupported array format {format}")

    def _convertible(self) -> object:
        """
        Get the data suitable for passing to an ``as_array`` method.
        """
        # look for a good format for the data. if we've already made a numpy,
        # use that; otherwise, try unknown, and fall back to torch (moved to
        # CPU). end result: convertible data.
        if self._numpy is not None:
            return self._numpy
        elif self._unknown is not None:
            return self._unknown
        elif self._torch is not None:
            return self._torch.cpu()
        else:
            raise RuntimeError("cannot find array data")

    def __array__(self, dtype=None, copy=None) -> NDArray[NPT]:
        arr = self.numpy()
        if dtype is not None:
            arr = np.require(arr, dtype)
        if copy:
            return arr.copy()
        else:
            return arr


MTFloatArray = MTArray[np.floating]
MTIntArray = MTArray[np.integer]
MTGenericArray = MTArray[np.generic]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
LensKit parallel computation support.
"""

from __future__ import annotations

from .config import effective_cpu_count, ensure_parallel_init, get_parallel_config, initialize
from .invoker import ModelOpInvoker, invoker
from .pool import multiprocess_executor

__all__ = [
    "initialize",
    "get_parallel_config",
    "effective_cpu_count",
    "ensure_parallel_init",
    "invoker",
    "ModelOpInvoker",
    "multiprocess_executor",
]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: basic
from __future__ import annotations

import logging
import os
import warnings
from dataclasses import dataclass
from typing import Optional

import torch
from threadpoolctl import threadpool_limits

_config: Optional[ParallelConfig] = None
_log = logging.getLogger(__name__)


@dataclass
class ParallelConfig:
    """
    Parallel processing configuration.
    """

    processes: int
    threads: int
    backend_threads: int
    child_threads: int

    @property
    def total_threads(self) -> int:
        """
        Get the total number of threads for this worker (training threads times
        backend threads).
        """
        return self.threads * self.backend_threads

    def env_vars(self) -> dict[str, str]:
        """
        Get this parallel configuration as a set of environment variables.  The
        set also includes ``OMP_NUM_THREADS`` to configure OMP early.
        """
        return {
            "LK_NUM_PROCS": str(self.processes),
            "LK_NUM_THREADS": str(self.threads),
            "LK_NUM_BACKEND_THREADS": str(self.backend_threads),
            "LK_NUM_CHILD_THREADS": str(self.child_threads),
        }


def initialize(
    config: ParallelConfig | None = None,
    *,
    processes: int | None = None,
    threads: int | None = None,
    backend_threads: int | None = None,
    child_threads: int | None = None,
):
    """
    Set up and configure LensKit parallelism.  This only needs to be called if
    you want to control when and how parallelism is set up; components using
    parallelism will call :func:`ensure_init`, which will call this function
    with its default arguments if it has not been called.

    .. seealso:: :ref:`parallel-config`

    Args:
        processes:
            The number of processes to use for multiprocessing evaluations (see
            :envvar:`LK_NUM_PROCS`)
        threads:
            The number of threads to use for parallel model training and similar
            operations (see :envvar:`LK_NUM_THREADS`).
        backend_threads:
            The number of threads underlying computational engines should use
            (see :envvar:`LK_NUM_BACKEND_THREADS`).
        child_threads:
            The number of threads backends are allowed to use in the worker
            processes in multiprocessing operations (see
            :envvar:`LK_NUM_CHILD_THREADS`).
    """
    global _config
    if _config:
        _log.warning("parallelism already initialized")
        return

    # our parallel computation doesn't work with FD sharing
    torch.multiprocessing.set_sharing_strategy("file_system")

    if config is None:
        _config = _resolve_parallel_config(processes, threads, backend_threads, child_threads)
    else:
        _config = config
    _log.debug("configuring for parallelism: %s", _config)

    threadpool_limits(_config.backend_threads, "blas")
    try:
        torch.set_num_interop_threads(_config.threads)
    except RuntimeError as e:
        _log.warning("failed to configure Pytorch interop threads: %s", e)
        warnings.warn("failed to set interop threads", RuntimeWarning)
    try:
        torch.set_num_threads(_config.backend_threads)
    except RuntimeError as e:
        _log.warning("failed to configure Pytorch intra-op threads: %s", e)
        warnings.warn("failed to set intra-op threads", RuntimeWarning)


def ensure_parallel_init():
    """
    Make sure LensKit parallelism is configured, and configure with defaults if
    it is not.

    Components using parallelism or intensive computations should call this
    function before they begin training.
    """
    if not _config:
        initialize()


def get_parallel_config() -> ParallelConfig:
    """
    Ensure that parallelism is configured and return the configuration.
    """
    ensure_parallel_init()
    assert _config is not None
    return _config


def subprocess_config(
    processes: int | None = None,
    threads: int | None = None,
    backend_threads: int | None = None,
    child_threads: int | None = None,
) -> ParallelConfig:
    """
    Get a parallel configuration for a subprocess.
    """
    cfg = get_parallel_config()
    return ParallelConfig(
        processes=processes or 1,
        threads=threads or 1,
        backend_threads=backend_threads or cfg.child_threads,
        child_threads=child_threads or 1,
    )


def effective_cpu_count() -> int:
    """
    Return the effective CPU count using the best available data.  Tries the following in order:

    1.  :func:`os.process_cpu_count`
    2.  :func:`os.sched_getaffinity`
    3.  :func:`os.cpu_count`
    """

    if hasattr(os, "process_cpu_count"):
        return os.process_cpu_count()  # type: ignore
    elif hasattr(os, "sched_getaffinity"):
        return len(os.sched_getaffinity(0))  # type: ignore
    else:
        ncpus = os.cpu_count()
        if ncpus is not None:
            return ncpus
        else:
            _log.warning("no CPU count available, assumping single CPU")
            return 1


def _resolve_parallel_config(
    processes: int | None = None,
    threads: int | None = None,
    backend_threads: int | None = None,
    child_threads: int | None = None,
) -> ParallelConfig:
    nprocs = os.environ.get("LK_NUM_PROCS", None)
    nthreads = os.environ.get("LK_NUM_THREADS", None)
    nbthreads = os.environ.get("LK_NUM_BACKEND_THREADS", None)
    cthreads = os.environ.get("LK_NUM_CHILD_THREADS", None)
    ncpus = effective_cpu_count()

    if processes is None and nprocs:
        processes = int(nprocs)

    if threads is None and nthreads:
        threads = int(nthreads)

    if backend_threads is None and nbthreads:
        backend_threads = int(nbthreads)

    if child_threads is None and cthreads:
        child_threads = int(cthreads)

    if processes is None:
        processes = min(ncpus, 4)

    if threads is None:
        threads = min(ncpus, 8)

    if backend_threads is None:
        backend_threads = max(min(ncpus // threads, 4), 1)

    if child_threads is None:
        child_threads = max(min(ncpus // processes, 4), 1)

    return ParallelConfig(processes, threads, backend_threads, child_threads)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import inspect
import os
import pickle
from time import perf_counter
from typing import ClassVar, Literal

import numpy as np
import pandas as pd

from pytest import approx, fixture, mark, skip

from lenskit import batch
from lenskit.data import Dataset, ItemList, RecQuery, from_interactions_df
from lenskit.data.builder import DatasetBuilder
from lenskit.pipeline import Component, topn_pipeline
from lenskit.splitting import split_temporal_fraction
from lenskit.training import Trainable, TrainingOptions

from ._markers import jit_enabled

retrain = os.environ.get("LK_TEST_RETRAIN")


class BasicComponentTests:
    component: type[Component]
    configs = []

    def test_instantiate_default(self):
        inst = self.component()
        assert inst is not None

        if self.component.config_class() is not None:
            assert inst.config is not None
        else:
            assert inst.config is None

    def test_default_config_vars(self):
        inst = self.component()
        cfg = inst.dump_config()
        for name, value in cfg.items():
            assert hasattr(inst.config, name)

    def test_default_config_round_trip(self):
        inst = self.component()
        cfg = inst.dump_config()

        i2 = self.component(self.component.validate_config(cfg))
        assert i2 is not inst
        assert isinstance(i2, self.component)
        print(cfg)
        print(i2.dump_config())
        assert i2.dump_config() == cfg

    def test_config_round_trip(self):
        if not self.configs:
            skip("no test configs specified")

        for cfg in self.configs:
            inst = self.component(self.component.validate_config(cfg))
            c1 = inst.dump_config()

            i2 = self.component(self.component.validate_config(c1))
            c2 = i2.dump_config()
            # config may be changed from source (due to normalization), but should
            # round-trip.
            assert c2 == c1


class TrainingTests:
    """
    Common tests for component training.
    """

    needs_jit: ClassVar[bool] = True
    component: type[Component]

    def maybe_skip_nojit(self):
        if self.needs_jit and not jit_enabled:
            skip("JIT is disabled")

    @fixture(scope="function" if retrain else "class")
    def trained_model(self, ml_ds: Dataset):
        self.maybe_skip_nojit()

        model = self.component()
        if isinstance(model, Trainable):
            model.train(ml_ds, TrainingOptions())
        yield model

    def test_skip_retrain(self, ml_ds: Dataset):
        self.maybe_skip_nojit()
        model = self.component()
        if not isinstance(model, Trainable):
            skip(f"component {model.__class__.__name__} is not trainable")

        model.train(ml_ds, TrainingOptions())
        v1_data = pickle.dumps(model)

        # train again
        t1 = perf_counter()
        model.train(ml_ds, TrainingOptions(retrain=False))
        t2 = perf_counter()
        # that should be very fast, let's say 10ms
        assert t2 - t1 < 0.01
        # the model shouldn't have changed
        v2_data = pickle.dumps(model)
        assert v2_data == v1_data


class ScorerTests(TrainingTests):
    """
    Common tests for scorer components.  Many of these just test that the component
    runs, not that it produces correct output.
    """

    component: type[Component]
    can_score: ClassVar[Literal["some", "known", "all"]] = "known"
    "What can this scorer score?"

    def invoke_scorer(self, inst: Component, **kwargs):
        sig = inspect.signature(inst)
        args = {n: v for (n, v) in kwargs.items() if n in sig.parameters}
        return inst(**args)

    def test_score_known(self, rng: np.random.Generator, ml_ds: Dataset, trained_model: Component):
        for u in rng.choice(ml_ds.users.ids(), 100):
            item_nums = rng.choice(ml_ds.item_count, 100, replace=False)
            items = ItemList(item_nums=item_nums, vocabulary=ml_ds.items)
            scored = self.invoke_scorer(trained_model, query=u, items=items)
            assert isinstance(scored, ItemList)
            assert np.all(scored.numbers() == item_nums)
            assert np.all(scored.ids() == items.ids())
            scores = scored.scores()
            assert scores is not None
            if self.can_score in ("known", "all"):
                assert np.all(np.isfinite(scores))

    def test_pickle_roundrip(
        self, rng: np.random.Generator, ml_ds: Dataset, trained_model: Component
    ):
        data = pickle.dumps(trained_model, pickle.HIGHEST_PROTOCOL)
        tm2 = pickle.loads(data)

        for u in rng.choice(ml_ds.users.ids(), 100):
            item_nums = rng.choice(ml_ds.item_count, 100, replace=False)
            items = ItemList(item_nums=item_nums, vocabulary=ml_ds.items)
            scored = self.invoke_scorer(trained_model, query=u, items=items)
            assert isinstance(scored, ItemList)

            s2 = self.invoke_scorer(tm2, query=u, items=items)
            assert isinstance(s2, ItemList)

            assert np.all(scored.numbers() == item_nums)
            assert np.all(scored.ids() == items.ids())

            arr = scored.scores()
            assert arr is not None
            arr2 = s2.scores()
            assert arr2 is not None
            assert arr == approx(arr2, nan_ok=True)

    def test_score_unknown_user(
        self, rng: np.random.Generator, ml_ds: Dataset, trained_model: Component
    ):
        "score with an unknown user ID"
        item_nums = rng.choice(ml_ds.item_count, 100, replace=False)
        items = ItemList(item_nums=item_nums, vocabulary=ml_ds.items)

        scored = self.invoke_scorer(trained_model, query=-1348, items=items)
        scores = scored.scores("pandas", index="ids")
        assert scores is not None
        if self.can_score == "all":
            assert np.all(np.isfinite(scores))

    def test_score_unknown_item(
        self, rng: np.random.Generator, ml_ds: Dataset, trained_model: Component
    ):
        "score with one target item unknown"
        item_nums = rng.choice(ml_ds.item_count, 100, replace=False)
        item_ids = ml_ds.items.ids(item_nums).tolist()
        item_ids.append(-318)
        items = ItemList(item_ids)

        scored = self.invoke_scorer(trained_model, query=ml_ds.users.id(0), items=items)
        scores = scored.scores("pandas", index="ids")
        assert scores is not None
        if self.can_score == "all":
            assert np.all(np.isfinite(scores))
        elif self.can_score == "known":
            assert np.all(np.isfinite(scores[:-1]))

    def test_score_empty_query(
        self, rng: np.random.Generator, ml_ds: Dataset, trained_model: Component
    ):
        "score with an empty query"
        item_nums = rng.choice(ml_ds.item_count, 100, replace=False)
        items = ItemList(item_nums=item_nums, vocabulary=ml_ds.items)
        q = RecQuery()
        scored = self.invoke_scorer(trained_model, query=q, items=items)
        assert np.all(scored.numbers() == item_nums)
        assert np.all(scored.ids() == items.ids())
        scores = scored.scores()
        assert scores is not None

    def test_score_query_history(
        self, rng: np.random.Generator, ml_ds: Dataset, trained_model: Component
    ):
        "score when query has user ID and history"
        u = rng.choice(ml_ds.users.ids())
        u_row = ml_ds.user_row(u)

        item_nums = rng.choice(ml_ds.item_count, 100, replace=False)
        items = ItemList(item_nums=item_nums, vocabulary=ml_ds.items)
        q = RecQuery(user_id=u, user_items=u_row)
        scored = self.invoke_scorer(trained_model, query=q, items=items)
        assert np.all(scored.numbers() == item_nums)
        assert np.all(scored.ids() == items.ids())
        scores = scored.scores()
        assert scores is not None

    def test_score_query_history_only(
        self, rng: np.random.Generator, ml_ds: Dataset, trained_model: Component
    ):
        "score when query only has history"
        u = rng.choice(ml_ds.users.ids())
        u_row = ml_ds.user_row(u)

        item_nums = rng.choice(ml_ds.item_count, 100, replace=False)
        items = ItemList(item_nums=item_nums, vocabulary=ml_ds.items)
        q = RecQuery(user_items=u_row)
        scored = self.invoke_scorer(trained_model, query=q, items=items)
        assert np.all(scored.numbers() == item_nums)
        assert np.all(scored.ids() == items.ids())
        scores = scored.scores()
        assert scores is not None

    def test_score_empty_items(
        self, rng: np.random.Generator, ml_ds: Dataset, trained_model: Component
    ):
        "score an empty list of items"
        u = rng.choice(ml_ds.users.ids())

        items = ItemList()
        scored = self.invoke_scorer(trained_model, query=u, items=items)
        assert len(scored) == 0
        scores = scored.scores()
        assert scores is not None

    def test_train_score_items_missing_data(self, rng: np.random.Generator, ml_ds: Dataset):
        "train and score when some entities are missing data"
        self.maybe_skip_nojit()
        drop_i = rng.choice(ml_ds.items.ids(), 20)
        drop_u = rng.choice(ml_ds.users.ids(), 5)

        dsb = DatasetBuilder(ml_ds)
        df = ml_ds.interactions().pandas(ids=True)
        df = df[~(df["user_id"].isin(drop_u))]
        df = df[~(df["item_id"].isin(drop_i))]

        iname = ml_ds.default_interaction_class()
        dsb.clear_relationships(iname)
        dsb.add_relationships(iname, df, entities=["user", "item"])
        ds = dsb.build()

        model = self.component()
        assert isinstance(model, Trainable)
        model.train(ds, TrainingOptions())

        good_u = rng.choice(ml_ds.users.ids(), 10, replace=False)
        for u in set(good_u) | set(drop_u):
            items = rng.choice(ml_ds.items.ids(), 50, replace=False)
            items = np.unique(np.concatenate([items, rng.choice(drop_i, 5)]))
            items = ItemList(items, vocabulary=ds.items)

            scored = self.invoke_scorer(model, query=u, items=items)
            assert len(scored) == len(items)
            assert np.all(scored.numbers() == items.numbers())
            assert np.all(scored.ids() == items.ids())

            scores = scored.scores()
            assert scores is not None

    @mark.slow
    def test_train_recommend(self, ml_ds: Dataset):
        """
        Test that a full train-recommend pipeline works.
        """
        self.maybe_skip_nojit()
        split = split_temporal_fraction(ml_ds, 0.2)
        model = self.component()
        pipe = topn_pipeline(model)
        pipe.train(split.train)

        recs = batch.recommend(pipe, split.test)
        assert len(recs) == len(split.test)

    @mark.slow
    def test_run_with_doubles(self, ml_ratings: pd.DataFrame):
        self.maybe_skip_nojit()
        ml_ratings = ml_ratings.astype({"rating": "f8"})
        ml_ds = from_interactions_df(ml_ratings)
        split = split_temporal_fraction(ml_ds, 0.3)
        model = self.component()
        pipe = topn_pipeline(model)
        pipe.train(split.train)

        recs = batch.recommend(pipe, split.test)
        assert len(recs) == len(split.test)



import warnings

from lenskit.logging import Stopwatch
from lenskit.random import derivable_rng, random_generator, set_global_rng

warnings.warn("lenskit.util is deprecated, import from original modules", DeprecationWarning)

__all__ = [
    "Stopwatch",
    "derivable_rng",
    "random_generator",
    "set_global_rng",
]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Utilities to manage randomness in LensKit and LensKit experiments.
"""

# pyright: strict
from __future__ import annotations

import os
from abc import abstractmethod
from hashlib import md5
from pathlib import Path
from typing import TYPE_CHECKING
from uuid import UUID

import numpy as np
from numpy.random import Generator, SeedSequence, default_rng
from typing_extensions import Any, Literal, Protocol, Sequence, TypeAlias, override

if TYPE_CHECKING:  # avoid circular import
    from lenskit.data import RecQuery

SeedLike: TypeAlias = int | Sequence[int] | np.random.SeedSequence
"""
Type for RNG seeds (see `SPEC 7`_).

.. _SPEC 7: https://scientific-python.org/specs/spec-0007/
"""

RNGLike: TypeAlias = np.random.Generator | np.random.BitGenerator
"""
Type for random number generators as inputs (see `SPEC 7`_).

.. _SPEC 7: https://scientific-python.org/specs/spec-0007/
"""

RNGInput: TypeAlias = SeedLike | RNGLike | None
"""
Type for RNG inputs (see `SPEC 7`_).

.. _SPEC 7: https://scientific-python.org/specs/spec-0007/
"""

ConfiguredSeed: TypeAlias = int | Sequence[int] | None
"""
Random number seed that can be configured.
"""

SeedDependency = Literal["user"]

_global_rng: Generator | None = None


DerivableSeed: TypeAlias = ConfiguredSeed | SeedDependency | tuple[ConfiguredSeed, SeedDependency]


def load_seed(file: Path | os.PathLike[str] | str, key: str = "random.seed") -> SeedSequence:
    """
    Load a seed from a configuration file.

    Args:
        file:
            The path to the configuration file.
        key:
            The path within the configuration object to the random seed.
    """
    data: Any
    file = Path(file)

    with file.open("rb") as f:
        match file.suffix:
            case ".toml":
                import tomllib

                data = tomllib.load(f)
            case ".yaml":
                import yaml

                data = yaml.load(f, yaml.SafeLoader)
            case ".json":
                import json

                data = json.load(f)
            case _:
                raise ValueError(f"unsupported config file {file.name}")

    parts = key.split(".")
    for part in parts:
        if isinstance(data, dict):
            data = data[part]
        else:
            raise TypeError(f"unsupported data type {type(data)}")

    return make_seed(data)


def set_global_rng(seed: RNGInput):
    """
    Set the global default RNG.
    """
    global _global_rng
    _global_rng = default_rng(seed)


def random_generator(seed: RNGInput = None) -> Generator:
    """
    Create a a random generator with the given seed, falling back to a global
    generator if no seed is provided.  If no global generator has been
    configured with :func:`set_global_rng`, it returns a fresh random RNG.
    """

    global _global_rng
    if seed is None and _global_rng is not None:
        return _global_rng
    else:
        return default_rng(seed)


def make_seed(
    *keys: SeedSequence | int | str | bytes | UUID | Sequence[int] | np.integer[Any] | None,
) -> SeedSequence:
    """
    Make an RNG seed from an input key, allowing strings as seed material.
    """
    seed: list[int] = []
    for key in keys:
        if key is None:
            continue
        elif isinstance(key, SeedSequence):
            ent = key.entropy
            if ent is None:
                continue
            elif isinstance(ent, int):
                seed.append(ent)
            else:
                seed += ent
        elif isinstance(key, np.integer):
            seed.append(key.item())
        elif isinstance(key, int):
            seed.append(key)
        elif isinstance(key, UUID):
            seed.append(_bytes_seed(key.bytes))
        elif isinstance(key, str):
            seed.append(_bytes_seed(key.encode("utf8")))
        elif isinstance(key, bytes):
            seed.append(_bytes_seed(key))
        elif isinstance(key, Sequence):  # type: ignore
            seed += key
        else:  # pragma: nocover
            raise TypeError(f"invalid key input: {key}")

    return SeedSequence(seed)


def _bytes_seed(key: bytes) -> int:
    digest = md5(key).digest()
    arr = np.frombuffer(digest, np.int32)
    return abs(np.bitwise_xor.reduce(arr).item())


class RNGFactory(Protocol):
    """
    Protocol for RNG factories that can do dynamic (e.g. per-user) seeding.
    """

    @abstractmethod
    def __call__(self, query: RecQuery | None) -> Generator:
        raise NotImplementedError()


class FixedRNG(RNGFactory):
    "RNG provider that always provides the same RNG"

    rng: Generator

    def __init__(self, rng: Generator):
        self.rng = rng

    @override
    def __call__(self, query: RecQuery | None = None) -> Generator:
        return self.rng

    def __str__(self):
        return "Fixed({})".format(self.rng)


class DerivingRNG(RNGFactory):
    "RNG provider that derives new RNGs from the key"

    seed: np.random.SeedSequence

    def __init__(self, seed: np.random.SeedSequence):
        self.seed = seed

    @override
    def __call__(self, query: RecQuery | None = None) -> Generator:
        if query is None or query.user_id is None:
            return np.random.default_rng(self.seed.spawn(1)[0])
        else:
            seed = make_seed(self.seed, query.user_id)  # type: ignore
            return default_rng(seed)

    def __str__(self):
        return "Derive({})".format(self.seed)


def derivable_rng(spec: DerivableSeed) -> RNGFactory:
    """
    RNGs that may be derivable from data in the query. These are for designs
    that need to be able to reproducibly derive RNGs for different keys, like
    user IDs (to make a “random” recommender produce the same sequence for the
    same user).

    Seed specifications may be any of the following:

    - A seed (:type:`~lenskit.random.SeedLike`).
    - The value ``'user'``, which will derive a seed from the query user ID.
    - A tuple of the form ``(seed, 'user')``, that will use ``seed`` as the
      basis and drive from it a new seed based on the user ID.

    .. seealso:: :ref:`rng`

    Args:
        spec:
            The seed specification.

    Returns:
        function:
            A function taking one (or more) key values, like
            :func:`derive_seed`, and returning a random number generator.
    """

    if spec == "user":
        return DerivingRNG(SeedSequence())
    elif isinstance(spec, tuple):
        seed, key = spec
        if key != "user":
            raise ValueError("unrecognized key %s", key)
        return DerivingRNG(make_seed(seed))
    else:
        return FixedRNG(default_rng(spec))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import math

import numpy as np
import torch
from typing_extensions import override

from lenskit.data import Dataset, ItemList
from lenskit.logging.progress import item_progress_handle, pbh_update
from lenskit.math.solve import solve_cholesky
from lenskit.parallel.chunking import WorkChunks

from ._common import ALSBase, ALSConfig, TrainContext, TrainingData


class ImplicitMFConfig(ALSConfig):
    weight: float = 40
    """
    The confidence weight for positive examples.
    """
    use_ratings: bool = False
    """
    If ``True``, use rating values instead of just presence or absence.
    """


class ImplicitMFScorer(ALSBase):
    """
    Implicit matrix factorization trained with alternating least squares
    :cite:p:`hu:implicit-mf`.  This algorithm outputs 'predictions', but they
    are not on a meaningful scale.  If its input data contains ``rating``
    values, these will be used as the 'confidence' values; otherwise, confidence
    will be 1 for every rated item.

    With weight :math:`w`, this function decomposes the matrix
    :math:`\\mathbb{1}^* + Rw`, where :math:`\\mathbb{1}^*` is an :math:`m
    \\times n` matrix of all 1s.

    See the base class :class:`ALSBase` for documentation on the estimated
    parameters you can extract from a trained model. See
    :class:`ImplicitMFConfig` and :class:`ALSConfig` for the configuration
    options for this component.

    .. versionchanged:: 2025.1
        ``ImplicitMFScorer`` no longer supports multiple training methods. It always uses
        Cholesky decomposition now.

    .. versionchanged:: 0.14
        By default, ``ImplicitMF`` ignores a ``rating`` column if one is present in the training
        data.  This can be changed through the ``use_ratings`` option.

    .. versionchanged:: 0.13
        In versions prior to 0.13, ``ImplicitMF`` used the rating column if it was present.
        In 0.13, we added an option to control whether or not the rating column is used; it
        initially defaulted to ``True``, but with a warning.  In 0.14 it defaults to ``False``.

    Stability:
        Caller
    """

    config: ImplicitMFConfig

    OtOr_: torch.Tensor

    @override
    def prepare_data(self, data: Dataset) -> TrainingData:
        if self.config.use_ratings:
            rmat = data.interaction_matrix(format="torch", field="rating")
        else:
            rmat = data.interaction_matrix(format="torch")

        rmat = torch.sparse_csr_tensor(
            crow_indices=rmat.crow_indices(),
            col_indices=rmat.col_indices(),
            values=rmat.values() * self.config.weight,
            size=rmat.shape,
        )
        return TrainingData.create(data.users, data.items, rmat)

    @override
    def initial_params(self, nrows: int, ncols: int, rng: np.random.Generator) -> torch.Tensor:
        mat = rng.standard_normal((nrows, ncols)) * 0.01
        mat = torch.from_numpy(mat)
        mat.square_()
        return mat

    @override
    def als_half_epoch(self, epoch: int, context: TrainContext) -> float:
        chunks = WorkChunks.create(context.nrows)

        OtOr = _implicit_otor(context.right, context.reg)
        with item_progress_handle(f"epoch {epoch} {context.label}s", total=context.nrows) as pbh:
            return _train_implicit_cholesky_fanout(context, OtOr, chunks, pbh)

    def finalize_training(self):
        # compute OtOr and save it on the model
        reg = self.config.user_reg
        self.OtOr_ = _implicit_otor(self.item_features_, reg)
        return True

    @override
    def new_user_embedding(
        self, user_num: int | None, user_items: ItemList
    ) -> tuple[torch.Tensor, None]:
        ri_idxes = user_items.numbers("torch", vocabulary=self.items_)

        ri_good = ri_idxes >= 0
        ri_it = ri_idxes[ri_good]
        if self.config.use_ratings:
            ratings = user_items.field("rating", "torch")
            if ratings is None:
                raise ValueError("no ratings in user items")
            ri_val = ratings[ri_good] * self.config.weight
        else:
            ri_val = torch.full((len(ri_good),), self.config.weight)

        ri_val = ri_val.to(self.item_features_.dtype)

        u_feat = self._train_new_row(ri_it, ri_val, self.item_features_, self.OtOr_)
        return u_feat, None

    def _train_new_row(
        self, items: torch.Tensor, ratings: torch.Tensor, i_embeds: torch.Tensor, OtOr: torch.Tensor
    ) -> torch.Tensor:
        """
        Train a single user row with new rating data.

        Args:
            items: the item IDs the user has rated
            ratings: the user's ratings for those items (when rating values are used)
            other: the item-feature matrix
            OtOr: the pre-computed regularization and background matrix.

        Returns:
            The user-feature vector.
        """
        self.logger.debug("learning new user row", n_items=len(items))

        # we can optimize by only considering the nonzero entries of Cu-I
        # this means we only need the corresponding matrix columns
        M = i_embeds[items, :]
        # Compute M^T (C_u-I) M, restricted to these nonzero entries
        MMT = (M.T * ratings) @ M
        # Build the matrix for solving
        A = OtOr + MMT
        # Compute RHS - only used columns (p_ui != 0) values needed
        y = i_embeds.T[:, items] @ (ratings + 1.0)
        # and solve
        x = solve_cholesky(A, y)

        return x


@torch.jit.script
def _implicit_otor(other: torch.Tensor, reg: float) -> torch.Tensor:
    nf = other.shape[1]
    regmat = torch.eye(nf)
    regmat *= reg
    Ot = other.T
    OtO = Ot @ other
    OtO += regmat
    return OtO


def _train_implicit_cholesky_rows(
    ctx: TrainContext, OtOr: torch.Tensor, start: int, end: int, pbh: str
) -> torch.Tensor:
    result = ctx.left[start:end, :].clone()

    for i in range(start, end):
        row = ctx.matrix[i]
        (n,) = row.shape
        if n == 0:
            continue

        cols = row.indices()[0]
        vals = row.values().type(ctx.left.type())

        # we can optimize by only considering the nonzero entries of Cu-I
        # this means we only need the corresponding matrix columns
        M = ctx.right[cols, :]
        # Compute M^T (C_u-I) M, restricted to these nonzero entries
        MMT = (M.T * vals) @ M
        # assert MMT.shape[0] == ctx.n_features
        # assert MMT.shape[1] == ctx.n_features
        # Build the matrix for solving
        A = OtOr + MMT
        # Compute RHS - only used columns (p_ui != 0) values needed
        # Cu is rates + 1 for the cols, so just trim Ot
        y = ctx.right.T[:, cols] @ (vals + 1.0)
        # and solve
        x = solve_cholesky(A, y)
        # assert len(uv) == ctx.n_features
        result[i - start, :] = x

        pbh_update(pbh, 1)

    return result


@torch.jit.script
def _train_implicit_cholesky_fanout(
    ctx: TrainContext, OtOr: torch.Tensor, chunks: WorkChunks, pbh: str
) -> float:
    results: list[tuple[int, int, torch.jit.Future[torch.Tensor]]] = []
    for start in range(0, ctx.nrows, chunks.chunk_size):
        end = min(start + chunks.chunk_size, ctx.nrows)
        results.append(
            (start, end, torch.jit.fork(_train_implicit_cholesky_rows, ctx, OtOr, start, end, pbh))  # type: ignore
        )

    sqerr = torch.tensor(0.0)
    for start, end, r in results:
        M = r.wait()
        diff = (ctx.left[start:end, :] - M).ravel()
        sqerr += torch.dot(diff, diff)
        ctx.left[start:end, :] = M

    return math.sqrt(sqerr.item())



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path
from urllib.request import urlopen

import click
from humanize import naturalsize

from lenskit.logging import get_logger

_log = get_logger(__name__)
ML_LOC = "http://files.grouplens.org/datasets/movielens/"


@click.command("fetch")
@click.option("--movielens", "source", flag_value="movielens", help="fetch MovieLens data")
@click.option("--force", is_flag=True, help="overwrite existing file")
@click.option("-D", "--data-dir", "dest", type=Path, help="directory for downloaded data")
@click.argument("name", nargs=-1)
def fetch(source: str | None, dest: Path | None, name: list[str], force: bool):
    """
    Convert data into the LensKit native format.
    """

    if dest is None:
        dest = Path()

    match source:
        case None:
            _log.error("no data source specified")
            raise click.UsageError("no data source specified")
        case "movielens":
            for n in name:
                fetch_movielens(n, dest, force)
        case _:
            raise ValueError(f"unknown data format {source}")


def fetch_movielens(name: str, base_dir: Path, force: bool):
    """
    Fetch a MovieLens dataset.  The followings names are recognized:

    . ml-100k
    . ml-1m
    . ml-10m
    . ml-20m
    . ml-25m
    . ml-latest
    . ml-latest-small

    Args:
        name:
            The name of the dataset.
        base_dir:
            The base directory into which data should be downloaded.
    """
    zipname = f"{name}.zip"
    zipfile = base_dir / zipname
    zipurl = ML_LOC + zipname

    log = _log.bind(source="movielens", name=name, dest=str(zipfile))

    if zipfile.exists():
        if force:
            log.warning("output file already exists, ovewriting")
        else:
            log.info("output file already exists")
            return

    log.debug("ensuring parent directory exists")
    base_dir.mkdir(exist_ok=True, parents=True)

    log.info("downloading MovieLens data set")
    with zipfile.open("wb") as zf:
        res = urlopen(zipurl)
        block = res.read(8 * 1024 * 1024)
        while len(block):
            _log.debug("received %d bytes", len(block))
            zf.write(block)
            block = res.read(8 * 1024 * 1024)

    log.info("downloaded %s", naturalsize(zipfile.stat().st_size, binary=True))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Vocabularies of IDs, tags, etc.
"""

# pyright: basic
from __future__ import annotations

from typing import Hashable, Iterable, Iterator, Literal, Sequence, overload

import numpy as np
import pandas as pd
import pyarrow as pa
from numpy.typing import ArrayLike, NDArray

from .types import ID, IDArray, IDSequence


class Vocabulary:
    """
    Vocabularies of entity identifiers for the LensKit data model.

    This class supports bidirectional mappings between key-like data and
    congiguous nonnegative integer indices.  Its key use is to facilitate the
    entity ID vocabularies in :class:`~lenskit.data.Dataset`, but it can also be
    used for things like item tags.

    IDs in a vocabulary must be unique.  Constructing a vocabulary with
    ``reorder=True`` ensures uniqueness (and sorts the IDs), but does not
    preserve the order of IDs in the original input.

    It is currently a wrapper around :class:`pandas.Index`, but this fact is not
    part of the stable public API.

    Args:
        keys:
            The IDs to put in the vocabulary.
        name:
            The vocabulary name (i.e. the entity class it stores IDs for).
        reorder:
            If ``True``, sort and deduplicate the IDs.  If ``False`` (the
            default), use the IDs as-is (assigning each to their position in the
            input sequence).

    Stability:
        Caller
    """

    name: str | None
    "The name of the vocabulary (e.g. “user”, “item”)."
    _index: pd.Index
    "The Pandas index implementing the vocabulary."

    def __init__(
        self,
        keys: IDSequence | pd.Index | Iterable[ID] | None = None,
        name: str | None = None,
        *,
        reorder: bool = True,
    ):
        self.name = name
        if keys is None:
            keys = pd.Index([], dtype=np.int64)
        elif isinstance(keys, pd.Index):
            pass
        elif isinstance(keys, np.ndarray) or isinstance(keys, list) or isinstance(keys, pd.Series):
            if reorder:
                keys = np.unique(keys)  # type: ignore
            keys = pd.Index(keys)  # type: ignore
        elif isinstance(keys, (pa.Array, pa.ChunkedArray)):
            keys = keys.drop_null()
            if reorder:
                keys = keys.unique().sort()
            keys = pd.Index(keys.to_numpy())
        else:
            keys = pd.Index(sorted(set(keys)))  # type: ignore

        if not keys.is_unique:
            raise ValueError("IDs must be unique")
        self._index = keys.rename(name + "_id") if name is not None else keys

    @property
    def index(self) -> pd.Index:
        """
        The vocabulary as a Pandas index.

        Stability:
            Internal
        """
        return self._index

    @property
    def size(self) -> int:
        "Current vocabulary size."
        return len(self._index)

    @overload
    def number(self, term: object, missing: Literal["error"] = "error") -> int: ...
    @overload
    def number(self, term: object, missing: Literal["none"] | None) -> int | None: ...
    def number(
        self, term: object, missing: Literal["error", "none"] | None = "error"
    ) -> int | None:
        "Look up the number for a vocabulary ID."
        try:
            num = self._index.get_loc(term)
            assert isinstance(num, int)
            return num
        except KeyError as e:
            if missing == "error":
                raise e
            else:
                return None

    def numbers(
        self, terms: Sequence[Hashable] | ArrayLike, missing: Literal["error", "negative"] = "error"
    ) -> np.ndarray[int, np.dtype[np.int32]]:
        "Look up the numbers for an array of terms or IDs."
        nums = np.require(self._index.get_indexer_for(terms), dtype=np.int32)
        if missing == "error" and np.any(nums < 0):
            raise KeyError()
        return nums

    def term(self, num: int) -> object:
        """
        Look up the term with a particular number.  Negative indexing is **not** supported.
        """
        if num < 0:
            raise IndexError("negative numbers not supported")
        return self._index[num]

    def terms(self, nums: list[int] | NDArray[np.integer] | pd.Series | None = None) -> IDArray:
        """
        Get a list of terms, optionally for an array of term numbers.

        Args:
            nums:
                The numbers (indices) for of terms to retrieve.  If ``None``,
                returns all terms.

        Returns:
            The terms corresponding to the specified numbers, or the full array
            of terms (in order) if ``nums=None``.
        """
        if nums is not None:
            nums = np.asarray(nums, dtype=np.int32)
            if np.any(nums < 0):
                raise IndexError("negative numbers not supported")
            return self._index[nums].values
        else:
            return self._index.values

    def id(self, num: int) -> object:
        "Alias for :meth:`term`  for greater readability for entity ID vocabularies."
        return self.term(num)

    def ids(self, nums: list[int] | NDArray[np.integer] | pd.Series | None = None) -> IDArray:
        "Alias for :meth:`terms` for greater readability for entity ID vocabularies."
        return self.terms(nums)

    def __eq__(self, other: Vocabulary) -> bool:  # noqa: F821
        if self is other:
            return True

        if self.name == other.name and len(self) == len(other):
            return np.all(self.index == other.index).item()

        return False

    def __contains__(self, key: object) -> bool:
        return key in self._index

    def __iter__(self) -> Iterator[object]:
        return iter(self._index.values)

    def __len__(self) -> int:
        return self.size

    def __array__(self, dtype=None) -> np.ndarray:
        return self._index.values.__array__(dtype)

    def __str__(self) -> str:
        if self.name:
            return f"Vocabulary of {self.size} {self.name} terms"
        else:
            return f"Vocabulary of {self.size} terms"

    def __repr__(self) -> str:
        if self.name:
            return f"<Vocabulary(name={self.name}, size={self.size})>"
        else:
            return f"<Vocabulary(size={self.size})>"



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Prediction accuracy metrics.  See :ref:`eval-predict-accuracy` for an overview
and instructions on using these metrics.
"""

from __future__ import annotations

import logging

import numpy as np
import pandas as pd
from numpy.typing import NDArray
from typing_extensions import Callable, Literal, TypeAlias, override

from lenskit.data import ItemList
from lenskit.data.adapt import ITEM_COMPAT_COLUMN, normalize_columns
from lenskit.data.types import AliasedColumn

from ._base import DecomposedMetric, ListMetric, Metric

_log = logging.getLogger(__name__)

MissingDisposition: TypeAlias = Literal["error", "ignore"]
ScoreArray: TypeAlias = NDArray[np.floating] | pd.Series
PredMetric: TypeAlias = Callable[[ScoreArray, ScoreArray], float]


class PredictMetric(Metric):
    """
    Extension to the metric function interface for prediction metrics.

    In addition to the general metric interface, predict metrics can be called
    with a single item list (or item list collection) that has both ``scores``
    and a ``rating`` field.

    Args:
        missing_scores:
            The action to take when a test item has not been scored.  The
            default throws an exception, avoiding situations where non-scored
            items are silently excluded from overall statistics.
        missing_truth:
            The action to take when no test items are available for a scored
            item. The default is to also to fail; if you are scoring a superset
            of the test items for computational efficiency, set this to
            ``"ignore"``.

    Stability:
        Caller
    """

    # predict metrics usually cannot fill in default values
    default = None
    missing_scores: MissingDisposition
    missing_truth: MissingDisposition

    def __init__(
        self,
        missing_scores: MissingDisposition = "error",
        missing_truth: MissingDisposition = "error",
    ):
        self.missing_scores = missing_scores
        self.missing_truth = missing_truth

    def align_scores(
        self, predictions: ItemList, truth: ItemList | None = None
    ) -> tuple[pd.Series[float], pd.Series[float]]:
        """
        Align prediction scores and rating values, applying the configured
        missing dispositions.  The result is two Pandas series, predictions and
        truth, that are aligned and checked for missing data in accordance with
        the configured options.
        """
        if isinstance(predictions, pd.DataFrame):
            df = normalize_columns(
                predictions, ITEM_COMPAT_COLUMN, AliasedColumn("score", ["prediction"])
            )
            predictions = ItemList.from_df(df)

        if not isinstance(predictions, ItemList):  # pragma: nocover
            raise TypeError(f"predictions must be ItemList, not {type(predictions)}")
        if truth is not None and not isinstance(truth, ItemList):  # pragma: nocover
            raise TypeError(f"truth must be ItemList, not {type(truth)}")

        pred_s = predictions.scores("pandas", index="ids")
        assert pred_s is not None, "item list does not have scores"
        if truth is not None:
            rate_s = truth.field("rating", "pandas", index="ids")
        else:
            rate_s = predictions.field("rating", "pandas", index="ids")
        assert rate_s is not None, "no ratings provided"
        pred_s, rate_s = pred_s.align(rate_s, join="outer")

        pred_m = pred_s.isna()
        rate_m = rate_s.isna()

        if self.missing_scores == "error" and (nbad := np.sum(pred_m & ~rate_m)):
            raise ValueError(f"missing scores for {nbad} truth items")

        if self.missing_truth == "error" and (nbad := np.sum(rate_m & ~pred_m)):
            raise ValueError(f"missing truth for {nbad} scored items")

        return pred_s, rate_s


class RMSE(PredictMetric, ListMetric, DecomposedMetric):
    """
    Compute RMSE (root mean squared error).  This is computed as:

    .. math::
        \\sum_{r_{ui} \\in R} \\left(r_{ui} - s(i|u)\\right)^2

    This metric does not do any fallbacks; if you want to compute RMSE with
    fallback predictions (e.g. usign a bias model when a collaborative filter
    cannot predict), generate predictions with
    :class:`~lenskit.basic.FallbackScorer`.

    Stability:
        Caller
    """

    @override
    def measure_list(self, predictions: ItemList, test: ItemList | None = None, /) -> float:
        ps, ts = self.align_scores(predictions, test)
        err = ps - ts
        err *= err
        return np.sqrt(np.mean(err))

    @override
    def compute_list_data(self, output, test):
        ps, ts = self.align_scores(output, test)
        err = ps - ts
        err *= err
        return np.sum(err), len(err)

    @override
    def extract_list_metric(self, metric):
        tot, n = metric
        if n > 0:
            return np.sqrt(tot / n)
        else:
            return np.nan

    @override
    def global_aggregate(self, values):
        tot_sqerr = 0.0
        tot_n = 0.0
        for t, n in values:
            tot_sqerr += t
            tot_n += n

        if tot_n > 0:
            return np.sqrt(tot_sqerr / tot_n)
        else:
            return np.nan


class MAE(PredictMetric, ListMetric, DecomposedMetric):
    """
    Compute MAE (mean absolute error).  This is computed as:

    .. math::
        \\sum_{r_{ui} \\in R} \\left|r_{ui} - s(i|u)\\right|

    This metric does not do any fallbacks; if you want to compute MAE with
    fallback predictions (e.g. usign a bias model when a collaborative filter
    cannot predict), generate predictions with
    :class:`~lenskit.basic.FallbackScorer`.

    Stability:
        Caller
    """

    @override
    def measure_list(self, predictions: ItemList, test: ItemList | None = None, /) -> float:
        ps, ts = self.align_scores(predictions, test)
        err = ps - ts
        return np.mean(np.abs(err)).item()

    @override
    def compute_list_data(self, output, test):
        ps, ts = self.align_scores(output, test)
        err = ps - ts
        return np.sum(np.abs(err)), len(err)

    @override
    def extract_list_metric(self, metric):
        tot, n = metric
        if n > 0:
            return tot / n
        else:
            return np.nan

    @override
    def global_aggregate(self, values):
        tot_err = 0.0
        tot_n = 0.0
        for t, n in values:
            tot_err += t
            tot_n += n

        if n > 0:
            return tot_err / tot_n
        else:
            return np.nan



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Support for parallelism with Ray.
"""

from __future__ import annotations

import base64
import itertools
import os
import pickle
from collections.abc import Callable, Iterable, Iterator
from platform import python_version
from typing import Any, Generic

from lenskit.logging import Task, get_logger
from lenskit.logging.worker import WorkerContext, WorkerLogConfig

from .config import (
    ParallelConfig,
    effective_cpu_count,
    ensure_parallel_init,
    get_parallel_config,
    subprocess_config,
)
from .invoker import A, InvokeOp, M, ModelOpInvoker, R

LK_PROCESS_SLOT = "lk_process"
_worker_parallel: ParallelConfig
_log = get_logger(__name__)


def ensure_cluster():
    import ray

    if not ray.is_initialized():
        _log.debug("Ray is not initialized, initializing")
        try:
            init_cluster()
        except ValueError as e:
            _log.debug("Ray initialization failed", exception=e)
            if "existing cluster" in str(e):
                _log.info("Ray cluster already started, reusing")
                _log.warn("existing Ray cluster may not properly throttle LensKit jobs")
                ray.init()
            else:
                raise e


def init_cluster(
    *,
    num_cpus: int | None = None,
    proc_slots: int | None = None,
    resources: dict[str, float] | None = None,
    worker_parallel: ParallelConfig | None = None,
    limit_slots: bool = True,
    **kwargs,
):
    """
    Initialize or connect to a Ray cluster, with the LensKit options.

    The resulting cluster can be used by an invoker, or it can be used directly.
    The Ray invoker uses batching, though, so it only works well with many small
    tasks.

    Args:
        num_cpus:
            The total number of CPUs to allow. Defaults to
            :fun:`effective_cpu_count`.
        proc_slots:
            The number of “process slots” for LensKit parallel operations.
            Defaults to the LensKit process count.  These slots are recorded as
            the ``lk_process`` resource on the Ray cluster.
        resources:
            Additional custom resources to register in the Ray cluster.
        worker_parallel:
            Parallel processing configuration for worker processes.  If
            ``None``, uses the default.
        limit_slots:
            ``False`` to disable the LensKit slot interface.
        kwargs:
            Other options to pass to :func:`ray.init`.

    Stability:
        Experimental
    """
    global _worker_parallel
    import ray

    if resources is None:
        resources = {}
    else:
        resources = resources.copy()

    cfg = get_parallel_config()
    if proc_slots is None:
        proc_slots = cfg.processes
    if limit_slots:
        resources = {LK_PROCESS_SLOT: proc_slots}
    else:
        resources = {}
    if num_cpus is None:
        num_cpus = effective_cpu_count()

    if worker_parallel is None:
        worker_parallel = subprocess_config()
    _worker_parallel = worker_parallel

    env = worker_parallel.env_vars().copy()
    wc = WorkerLogConfig.current()
    env["LK_LOG_CONFIG"] = base64.encodebytes(pickle.dumps(wc)).decode()

    runtime = ray.runtime_env.RuntimeEnv(env_vars=env)

    _log.info("starting Ray cluster")
    ray.init(num_cpus=num_cpus, resources=resources, runtime_env=runtime, **kwargs)


def inference_worker_cpus() -> int:
    return _worker_parallel.backend_threads


def training_worker_cpus() -> int:
    return _worker_parallel.total_threads


def ray_supported() -> bool:
    """
    Check if this Ray setup is supported by LensKit.
    """
    if python_version() < "3.12":
        return False
    else:
        return ray_available()


def ray_available() -> bool:
    """
    Check if Ray is available.
    """
    try:
        import ray  # noqa: F401

        return True
    except ImportError:
        return False


def ray_active() -> bool:
    """
    Query whether Ray is active.
    """
    if ray_available():
        import ray

        return ray.is_initialized()
    else:
        return False


def is_ray_worker() -> bool:
    """
    Determine whether the current process is running on a Ray worker.
    """
    # logic adapted from https://discuss.ray.io/t/how-to-know-if-code-is-running-on-ray-worker/15642
    if ray_active():
        import ray

        ctx = ray.get_runtime_context()
        return ctx.worker.mode == ray.WORKER_MODE
    else:
        return False


class RayOpInvoker(ModelOpInvoker[A, R], Generic[M, A, R]):
    function: InvokeOp[M, A, R]
    model_ref: Any

    def __init__(
        self,
        model: M,
        func: InvokeOp[M, A, R],
    ):
        _log.debug("persisting to Ray cluster")
        import ray

        if isinstance(model, ray.ObjectRef):
            self.model_ref = model
        else:
            self.model_ref = ray.put(model)
        self.function = func
        slots = {}
        if LK_PROCESS_SLOT in ray.cluster_resources():
            slots = {LK_PROCESS_SLOT: 1}
        else:
            _log.warning(f"cluster has no resource {LK_PROCESS_SLOT}")

        worker = ray.remote(_ray_invoke_worker)
        self.action = worker.options(num_cpus=inference_worker_cpus(), resources=slots)

    def map(self, tasks: Iterable[A]) -> Iterator[R]:
        import ray

        batch_results = [
            self.action.remote(self.function, self.model_ref, batch)
            for batch in itertools.batched(tasks, 200)
        ]
        for br in batch_results:
            yield from ray.get(br)

    def shutdown(self):
        del self.model_ref


def _ray_invoke_worker(func: Callable[[M, A], R], model: M, args: list[A]) -> list[R]:
    log_cfg = pickle.loads(base64.decodebytes(os.environb[b"LK_LOG_CONFIG"]))
    ensure_parallel_init()
    with WorkerContext(log_cfg) as ctx:
        try:
            with Task("cluster worker", subprocess=True) as task:
                result = [func(model, arg) for arg in args]
        finally:
            ctx.send_task(task)

    return result



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Timing support
"""

import time


class Stopwatch:
    """
    Timer class for recording elapsed wall time in operations.
    """

    start_time = None
    stop_time = None

    def __init__(self, start=True):
        if start:
            self.start()

    def start(self):
        self.start_time = time.perf_counter()

    def stop(self):
        self.stop_time = time.perf_counter()

    def elapsed(self) -> float:
        assert self.start_time is not None
        stop = self.stop_time or time.perf_counter()

        return stop - self.start_time

    def __str__(self):
        elapsed = self.elapsed()
        if elapsed < 1:
            return "{: 0.0f}ms".format(elapsed * 1000)
        elif elapsed > 60 * 60:
            h, m = divmod(elapsed, 60 * 60)
            m, s = divmod(m, 60)
            return "{:0.0f}h{:0.0f}m{:0.2f}s".format(h, m, s)
        elif elapsed > 60:
            m, s = divmod(elapsed, 60)
            return "{:0.0f}m{:0.2f}s".format(m, s)
        else:
            return "{:0.2f}s".format(elapsed)

    def __repr__(self):
        elapsed = self.elapsed()
        if self.stop_time:
            return "<Stopwatch stopped at {:.3f}s>".format(elapsed)
        else:
            return "<Stopwatch running at {:.3f}s>".format(elapsed)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
from typing_extensions import override

from lenskit.data.items import ItemList

from ._base import ListMetric, RankingMetricBase


class RBP(ListMetric, RankingMetricBase):
    """
    Evaluate recommendations with rank-biased precision :cite:p:`rbp` with a
    patience parameter :math:`\\gamma`.

    If :math:`r_{ui} \\in \\{0, 1\\}` is binary implicit ratings, this is
    computed by:

    .. math::
        \\begin{align*}
        \\operatorname{RBP}_\\gamma(L, u) & =(1 - \\gamma) \\sum_i r_{ui} p^i
        \\end{align*}

    The original RBP metric depends on the idea that the rank-biased sum of
    binary relevance scores in an infinitely-long, perfectly-precise list has is
    :math:`1/(1 - \\gamma)`. However, in recommender evaluation, we usually have
    a small test set, so the maximum achievable RBP is significantly less, and
    is a function of the number of test items.  With ``normalize=True``, the RBP
    metric will be normalized by the maximum achievable with the provided test
    data.

    :cite:t:`rbp` provide an extended discussion on choosing the patience
    parameter :math:`\\gamma`.  This metric defaults to :math:`\\gamma=0.85`, to
    provide a relatively shallow curve and reward good items on the first few
    pages of results (in a 10-per-page setting).  Recommendation systems data
    has no pooling, so the variance of this estimator may be high as they note
    in the paper; however, RBP with high patience should be no worse than nDCG
    (and perhaps even better) in this regard.

    .. warning::

        The additional normalization is experimental, and should not yet be used
        for published research results.

    Args:
        k:
            The maximum recommendation list length.
        patience:
            The patience parameter :math:`\\gamma`, the probability that the
            user continues browsing at each point.  The default is 0.85.
        normalize:
            Whether to normalize the RBP scores; if ``True``, divides the RBP
            score by the maximum achievable with the test data (as in nDCG).

    Stability:
        Caller
    """

    patience: float
    normalize: bool

    def __init__(self, k: int | None = None, *, patience: float = 0.85, normalize: bool = False):
        super().__init__(k)
        self.patience = patience
        self.normalize = normalize

    @property
    def label(self):
        if self.k is not None:
            return f"RBP@{self.k}"
        else:
            return "RBP"

    @override
    def measure_list(self, recs: ItemList, test: ItemList) -> float:
        recs = self.truncate(recs)
        k = len(recs)

        nrel = len(test)
        if nrel == 0:
            return np.nan

        items = recs.ids()
        good = np.isin(items, test.ids())
        # γ^(r-1)
        disc = np.power(self.patience, np.arange(k))
        rbp = np.sum(disc[good])
        if self.normalize:
            # normalize by max achieveable RBP
            max = np.sum(disc[: min(nrel, k)])
            return rbp / max
        else:
            # standard RBP normalization
            return rbp * (1 - self.patience)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Recommender systems toolkit.
"""

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "UNSPECIFIED"


from .operations import predict, recommend, score
from .pipeline import Pipeline, RecPipelineBuilder, topn_pipeline

__all__ = ["predict", "recommend", "score", "Pipeline", "RecPipelineBuilder", "topn_pipeline"]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections import namedtuple
from collections.abc import Sequence

from typing_extensions import (
    Any,
    NamedTuple,
    TypeAlias,
    TypeVar,
    overload,
)

from ..types import ID

GenericKey: TypeAlias = tuple[ID, ...]
K = TypeVar("K", bound=tuple, default=GenericKey)
"""
Fixed key type.
"""
Ko = TypeVar("Ko", bound=tuple, default=GenericKey)
"""
Fixed key type for an "other" collection.
"""
KL = TypeVar("KL", covariant=True, bound=tuple, default=GenericKey)
"""
Covariant key type for key lookup.
"""
KeySchema: TypeAlias = type[K] | tuple[str, ...]


class UserIDKey(NamedTuple):
    """
    Key type for user IDs.  This is used for :class:`item list collections
    <ItemListCollection>` that are keyed by user ID, a common setup for
    recommendation runs and
    """

    user_id: ID


KEY_CACHE: dict[tuple[str, ...], type[tuple]] = {("user_id",): UserIDKey}


def key_fields(kt: type[tuple]) -> tuple[str]:
    "extract the fields from a key type"
    return kt._fields  # type: ignore


def key_dict(kt: tuple[ID, ...]) -> dict[str, Any]:
    return kt._asdict()  # type: ignore


@overload
def create_key(kt: type[K], *values: ID) -> K: ...
@overload
def create_key(kt: Sequence[str], *values: ID) -> GenericKey: ...
def create_key(kt: type[K] | Sequence[str], *values: ID) -> tuple[Any, ...]:
    if isinstance(kt, type):
        return kt(*values)  # type: ignore
    else:
        kt = create_key_type(*kt)  # type: ignore
        return kt(*values)  # type: ignore


def create_key_type(*fields: str) -> type[GenericKey]:
    """
    Create a new key
    """
    assert isinstance(fields, tuple)
    kt = KEY_CACHE.get(fields, None)
    if kt is None:
        ktn = f"LKILCKeyType{len(KEY_CACHE)+1}"
        kt = namedtuple(ktn, fields)
        # support pickling
        kt.__reduce__ = _reduce_generic_key  # type: ignore
        KEY_CACHE[fields] = kt
    return kt


def _reduce_generic_key(key):
    args = (key._fields,) + key
    return create_key, args


def project_key(key: tuple, target: type[K]) -> K:
    """
    Project a key onto a subset of its fields.  This is to enable keys to be
    looked up in other collections that are keyed on a subset of their fields,
    such as using a key consisting of a user ID and a sequence number to look up
    test data in a collection keyed only by user ID.
    """

    if isinstance(key, target):
        return key

    try:
        return target._make(getattr(key, f) for f in target._fields)  # type: ignore
    except AttributeError as e:
        raise TypeError(f"source key is missing field {e.name}")
